from pydantic import ValidationError
import numpy as np
import os
import pytest
import tempfile

from paretobench.containers import Experiment, Population, History


@pytest.mark.parametrize("generate_names", [False, True])
def test_experiment_save_load(generate_names):
    """
    Make a randomized experiment, save it to disk, load it, and then confirm everything matches.
    """
    # Create a randomized Experiment object
    experiment = Experiment.from_random(
        n_histories=32,
        n_populations=10,
        n_objectives=5,
        n_decision_vars=30,
        n_constraints=2,
        pop_size=50,
        generate_names=generate_names,
    )

    # Use a temporary directory to save the file
    with tempfile.TemporaryDirectory() as tmpdir:
        # Define the file path
        file_path = os.path.join(tmpdir, "test.h5")
        experiment.save(file_path)

        # Load the experiment from the file and compare with original
        loaded_experiment = Experiment.load(file_path)
        assert experiment == loaded_experiment, "The loaded experiment is not equal to the original experiment."


def test_empty_history():
    """
    Confirms save/load methods work correctly whan an empty history object is saved.
    """
    # Create an Experiment object
    experiment = Experiment(name="", runs=[History(reports=[], problem="")])

    # Use a temporary directory to save the file
    with tempfile.TemporaryDirectory() as tmpdir:
        # Define the file path
        file_path = os.path.join(tmpdir, "test.h5")
        experiment.save(file_path)

        # Load the experiment from the file and compare with original
        loaded_experiment = Experiment.load(file_path)
        assert len(experiment.runs[0]) == 0
        assert experiment == loaded_experiment, "The loaded experiment is not equal to the original experiment."


def test_generate_population():
    pop = Population(f=np.random.random((128, 3)), fevals=0)
    assert pop.x.shape[0] == 128
    assert pop.g.shape[0] == 128


@pytest.mark.parametrize(
    "f, g, idx",
    [
        (np.array([[1.0, 0.0], [0.0, 1.0]]).T, None, [True, True]),
        (np.array([[1.0, 0.0], [1.0, 1.0]]).T, None, [False, True]),
        (np.array([[0.0, 1.0], [1.0, 1.0]]).T, None, [True, False]),
        (np.array([[0.0, 1.0], [1.0, 0.0]]).T, np.array([[1.0, 0.1]]).T, [True, True]),
        (
            np.array([[0.0, 1.0], [1.0, 0.0]]).T,
            np.array([[1.0, -0.1]]).T,
            [True, False],
        ),
        (
            np.array([[0.0, 1.0], [1.0, 0.0]]).T,
            np.array([[-1.0, -0.1]]).T,
            [False, True],
        ),
        (
            np.array([[0.0, 1.0], [1.0, 0.0]]).T,
            np.array([[-1.0, -0.1], [100.0, -0.1], [100.0, 3e3]]).T,
            [False, True],
        ),
    ],
)
def test_get_nondominated_indices(f, g, idx):
    """
    Check domination function for Population on some simple examples.
    """
    pop = Population(f=f, g=g, fevals=0)
    assert all(pop.get_nondominated_indices() == idx)


def test_to_nondominated():
    """
    Confirm that History.to_nondominated produces the nondominated individuals up to each point in the runs.
    """
    # Make a mock history object
    hist = History.from_random(10, 3, 4, 0, 50)
    hist_nd = hist.to_nondominated()

    for report_idx, report_nd in enumerate(hist_nd.reports):
        # Calculate the nondominated individuals ourself
        ref_nd = sum(hist.reports[1 : report_idx + 1], hist.reports[0]).get_nondominated_set()

        # Sort test and reference set so they are comparable
        ref_idx = np.lexsort(np.concatenate((ref_nd.x, ref_nd.f), axis=1).T)
        test_idx = np.lexsort(np.concatenate((report_nd.x, report_nd.f), axis=1).T)

        # Make sure the two are the same
        np.testing.assert_array_equal(ref_nd.x[ref_idx, :], report_nd.x[test_idx, :])

        # Double check that fevals hasn't changed
        assert report_nd.fevals == hist.reports[report_idx].fevals


def test_population_batch_dimension():
    """
    Confirm validation of batch dimension size works correctly.
    """
    # Create valid arrays with matching batch dimensions
    valid_x = np.random.rand(10, 5)
    valid_f = np.random.rand(10, 3)
    valid_g = np.random.rand(10, 2)

    # Create invalid arrays with different batch dimensions
    invalid_x = np.random.rand(10, 5)
    invalid_f = np.random.rand(8, 3)
    invalid_g = np.random.rand(10, 2)

    # Test that creating a valid Population instance does not raise an error
    try:
        Population(x=valid_x, f=valid_f, g=valid_g, fevals=1)
    except ValidationError:
        pytest.fail("Population creation with valid batch dimensions raised ValidationError unexpectedly!")

    # Test that creating an invalid Population instance raises a ValidationError
    with pytest.raises(
        ValidationError,
        match=r".*Batch dimensions do not match \(len\(x\)=10, len\(f\)=8, len\(g\)=10\).*",
    ):
        Population(x=invalid_x, f=invalid_f, g=invalid_g, fevals=1)


def test_history_validation():
    """
    Make sure that consistency checks for history objects work.
    """
    # Create valid populations with consistent decision variables, objectives, and constraints
    valid_population_1 = Population.from_random(
        n_objectives=3, n_decision_vars=5, n_constraints=2, pop_size=10, fevals=1
    )
    valid_population_2 = Population.from_random(
        n_objectives=3, n_decision_vars=5, n_constraints=2, pop_size=10, fevals=2
    )

    # Create invalid populations
    invalid_population_decision_vars = Population.from_random(
        n_objectives=3, n_decision_vars=6, n_constraints=2, pop_size=10, fevals=3
    )
    invalid_population_objectives = Population.from_random(
        n_objectives=4, n_decision_vars=5, n_constraints=2, pop_size=10, fevals=4
    )
    invalid_population_constraints = Population.from_random(
        n_objectives=3, n_decision_vars=5, n_constraints=3, pop_size=10, fevals=5
    )

    # Test that creating a valid History instance does not raise an error
    try:
        History(
            reports=[valid_population_1, valid_population_2],
            problem="Test Problem",
            metadata={"description": "A valid test case"},
        )
    except ValidationError:
        pytest.fail("History creation with consistent populations raised ValidationError unexpectedly!")

    # Test that creating an invalid History instance raises a ValidationError due to inconsistent decision variables
    with pytest.raises(ValidationError, match="Inconsistent number of decision variables in reports"):
        History(
            reports=[valid_population_1, invalid_population_decision_vars],
            problem="Test Problem",
            metadata={"description": "An invalid test case with inconsistent decision variables"},
        )

    # Test that creating an invalid History instance raises a ValidationError due to inconsistent objectives
    with pytest.raises(ValidationError, match="Inconsistent number of objectives in reports"):
        History(
            reports=[valid_population_1, invalid_population_objectives],
            problem="Test Problem",
            metadata={"description": "An invalid test case with inconsistent objectives"},
        )

    # Test that creating an invalid History instance raises a ValidationError due to inconsistent constraints
    with pytest.raises(ValidationError, match="Inconsistent number of constraints in reports"):
        History(
            reports=[valid_population_1, invalid_population_constraints],
            problem="Test Problem",
            metadata={"description": "An invalid test case with inconsistent constraints"},
        )

    # Test for inconsistent names - case where some have names and others don't
    population_with_names = Population.from_random(
        n_objectives=3,
        n_decision_vars=5,
        n_constraints=2,
        pop_size=10,
        fevals=6,
    )
    population_with_names.names_x = ["var1", "var2", "var3", "var4", "var5"]
    population_with_names.names_f = ["obj1", "obj2", "obj3"]
    population_with_names.names_g = ["con1", "con2"]
    population_without_names = Population.from_random(
        n_objectives=3,
        n_decision_vars=5,
        n_constraints=2,
        pop_size=10,
        fevals=7,
    )

    with pytest.raises(ValidationError, match="Inconsistent names for decision variables in reports"):
        History(
            reports=[population_with_names, population_without_names],
            problem="Test Problem",
            metadata={"description": "An invalid test case with inconsistent names"},
        )

    # Test for inconsistent names - case where names are different across populations
    population_with_different_names = Population.from_random(
        n_objectives=3,
        n_decision_vars=5,
        n_constraints=2,
        pop_size=10,
        fevals=8,
    )
    population_with_different_names.names_x = ["varA", "varB", "varC", "varD", "varE"]
    population_with_different_names.names_f = ["obj1", "obj2", "obj3"]
    population_with_different_names.names_g = ["con1", "con2"]

    with pytest.raises(ValidationError, match="Inconsistent names for decision variables in reports"):
        History(
            reports=[population_with_names, population_with_different_names],
            problem="Test Problem",
            metadata={"description": "An invalid test case with inconsistent names"},
        )


def test_population_invalid_dimensions():
    batch_size = 5
    num_variables = 3

    # Invalid dimension (3D instead of 2D)
    x = np.random.rand(batch_size, num_variables, 2)
    f = np.random.rand(batch_size, 2)
    g = np.random.rand(batch_size, 1)

    # Expect ValueError for incorrect number of dimensions
    with pytest.raises(ValueError, match="Expected array with 2 dimensions for field 'x'"):
        Population(x=x, f=f, g=g, fevals=5)


def test_overwrite():
    # Create a randomized Experiment object
    experiment1 = Experiment.from_random(
        n_histories=32,
        n_populations=10,
        n_objectives=5,
        n_decision_vars=30,
        n_constraints=2,
        pop_size=50,
    )

    # A second experiment
    experiment2 = Experiment.from_random(
        n_histories=32,
        n_populations=10,
        n_objectives=5,
        n_decision_vars=30,
        n_constraints=2,
        pop_size=50,
    )

    with tempfile.TemporaryDirectory() as dir:
        # Overwrite the file
        experiment1.save(os.path.join(dir, "experiment.h5"))
        experiment2.save(os.path.join(dir, "experiment.h5"))

        # Load the experiment from the file and compare with original
        loaded_experiment = Experiment.load(os.path.join(dir, "experiment.h5"))
        assert experiment2 == loaded_experiment, "The loaded experiment is not equal to the original experiment."


def test_count_unique_individuals():
    # Test empty population
    pop_empty = Population(x=np.empty((0, 2)), f=np.empty((0, 1)), g=np.empty((0, 1)))
    assert pop_empty.count_unique_individuals() == 0

    # Test all identical individuals
    pop_identical = Population(
        x=np.array([[1.0, 2.0], [1.0, 2.0], [1.0, 2.0]]),
        f=np.array([[3.0], [3.0], [3.0]]),
        g=np.array([[4.0], [4.0], [4.0]]),
    )
    assert pop_identical.count_unique_individuals() == 1

    # Test all unique individuals
    pop_unique = Population(
        x=np.array([[1.0, 2.0], [2.0, 3.0], [3.0, 4.0]]),
        f=np.array([[3.0], [4.0], [5.0]]),
        g=np.array([[4.0], [5.0], [6.0]]),
    )
    assert pop_unique.count_unique_individuals() == 3

    # Test floating point tolerance
    pop_tolerance = Population(
        x=np.array([[1.0, 2.0], [1.0 + 1e-7, 2.0 - 1e-7], [1.5, 2.5]]),
        f=np.array([[3.0], [3.0 + 1e-7], [3.5]]),
        g=np.array([[4.0], [4.0 - 1e-7], [4.5]]),
    )
    assert pop_tolerance.count_unique_individuals(decimals=4) == 2
    assert pop_tolerance.count_unique_individuals() == 3

    # Test with empty dimensions
    pop_empty_dims = Population(
        x=np.empty((3, 0)),
        f=np.array([[1.0], [2.0], [1.0]]),
        g=np.array([[3.0], [4.0], [3.0]]),
    )
    assert pop_empty_dims.count_unique_individuals() == 2



from .exceptions import SerializationError, DeserializationError


def dumps(data: dict):
    """Converts a dict of basic python types (float, int, string, bool) to a single line string. This method is used to help
    instantiate problem objects with parameters from single line strings (such as those in config files).

    Parameters
    ----------
    data : dict
        The data to serialize

    Returns
    -------
    str
        The single line containing the data

    Raises
    ------
    SerializationError
        Object could not be converted
    """
    # Confirm object is appropriate for the function
    for k, v in data.items():
        # Check for illegal characters in the keys
        if not k.replace("_", "").isalnum():
            raise SerializationError(f'Keys must only contain alphanumeric characters. Found key named "{k}"')

        # Check for illegal datatypes
        if type(v) not in [int, float, bool, str]:
            raise SerializationError(f'Cannot serialize object of type {type(v)} at key "{k}"')

    # Perform the conversion
    items = []
    for k, v in data.items():
        if isinstance(v, str):
            # Escape problem characters before writing
            v_safe = v.replace("\\", "\\\\")
            v_safe = v_safe.replace('"', r"\"")
            items.append(f'{k}="{v_safe}"')
        else:
            items.append(f"{k}={v}")
    return ", ".join(sorted(items))


def split_unquoted(s: str, split_char=",", quote_char='"', escape_char="\\"):
    """Breaks string into pieces by the character `split_char` as long as it isn't in a quoted section. Quotes may be escaped
    using a user provided escape character.

    Parameters
    ----------
    s : str
        The string to split up
    split_char : str, optional
        The character to split on, by default ','
    quote_char : str, optional
        The character recognized as a quote, by default '"'
    escape_char : str, optional
        The character recognized as an escape character, by default '\\'

    Raises
    ------
    DeserializationError
        Detected a badly formed string. For instance unclosed quotes or escape character in a weird spot.
    """
    # Note: tried to replace with csv.reader, but it only supports quotes around the entire field separated by commas
    # Go through char by char while keeping track of "depth" into quotes. Store chunks of str as we go
    last_escape_char = -2  # The index to the last escape character seen
    in_quotes = False  # Are we currently in quotes?
    chunk_start = 0  # The start of the current chunk
    chunks = []  # The list of chunks we have seen
    for idx, c in enumerate(s):
        # We saw a non-escaped quote marks
        if c == quote_char and (last_escape_char != idx - 1):
            in_quotes = not in_quotes

        # We saw the splitter character outside of quotes
        elif not in_quotes and c == split_char:
            chunks.append(s[chunk_start:idx])
            chunk_start = idx + 1

        # We saw the escape character
        elif c == escape_char:
            last_escape_char = idx
            if not in_quotes:
                raise DeserializationError("Detected escaped character outside of a string. Possible corrupted data.")

    # If we didn't end on a split_char, add remaining values
    if chunk_start < len(s):
        chunks.append(s[chunk_start:])

    # Check that we ended with a quote
    if in_quotes:
        raise DeserializationError("Unterminated quotations marks. Possible corrupted data.")

    return chunks


def can_convert_to_int(x):
    """
    Whether or not the "raw value" string used in the deserialization function `loads(...)` can be converted to an int.

    Parameters
    ----------
    x : str
        The string under test

    Returns
    -------
    bool
        Whether or not the string can be converted
    """
    try:
        int(x)
        return True
    except ValueError:
        return False


def can_convert_to_float(x):
    """
    Whether or not the "raw value" string used in the deserialization function `loads(...)` can be converted to a float.

    Parameters
    ----------
    x : str
        The string under test

    Returns
    -------
    bool
        Whether or not the string can be converted
    """
    try:
        float(x)
        return True
    except ValueError:
        return False


def can_convert_to_bool(x):
    """
    Whether or not the "raw value" string used in the deserialization function `loads(...)` can be converted to a bool.

    Parameters
    ----------
    x : str
        The string under test

    Returns
    -------
    bool
        Whether or not the string can be converted
    """
    return x in ["True", "False"]


def is_quoted_str(x):
    """
    Is the string x surrounded by quotes. Used to determine whether or not the string represents a string-type value in the
    serialization format.

    Parameters
    ----------
    x : str
        The string to be parsed

    Returns
    -------
    bool
        Whether or not x should be treated as a string
    """
    return x[0] == '"'


def loads(s: str):
    """Converts a string in the "single line" serialization format back to a dict of basic python objects.

    Parameters
    ----------
    s : str
        The string containing data in "single line" format

    Returns
    -------
    dict
        The dict containing data loaded from the string

    Raises
    ------
    DeserializationError
        Deserializing the data failed
    """
    # Clean all whitespace around the string
    s = s.strip()

    # Break string into sections
    ret = {}
    for chunk in split_unquoted(s):
        # Make sure we can break out the key and value
        if "=" not in chunk:
            raise DeserializationError(f'Bad key value pair: "{chunk}"')

        # Break into key and value
        idx = chunk.index("=")  # Must use first instance in case '=' in a string object
        k = chunk[:idx].strip()
        v_raw = chunk[idx + 1 :].strip()

        # Handle strings
        if is_quoted_str(v_raw):
            # Replace escaped characters and get rid of surrounding quotes
            v_raw = v_raw.replace("\\\\", "\\")
            v_raw = v_raw.replace('\\"', '"')
            v_parsed = v_raw[1:-1]
        elif can_convert_to_bool(v_raw):
            v_parsed = {"True": True, "False": False}[v_raw]
        elif can_convert_to_int(v_raw):
            v_parsed = int(v_raw)
        elif can_convert_to_float(v_raw):
            v_parsed = float(v_raw)
        else:
            raise DeserializationError(
                f'Could not determine data type for the value associated with key "{k}": "{v_raw}"'
            )

        # Finally set the dict entry
        ret[k] = v_parsed
    return ret



import numpy as np
import pandas as pd
import pandas.api.types as ptypes
import pytest

from paretobench import eval_metrics_experiments, Problem
from paretobench.analyze_metrics import (
    aggregate_metric_series,
    construct_metric_comparison_table,
    comparison_table_to_latex,
)
from paretobench.analyze_metrics import (
    aggregate_metrics_feval_budget,
    apply_feval_cutoff,
    normalize_problem_name,
)
from .utils import generate_moga_experiments, example_metric


def test_aggregate_metrics_feval_budget():
    """
    Test the aggregate function. Confirm that means and stds are taken correctly and the index data is carried through
    correctly.
    """
    # Create some test objects
    runs = generate_moga_experiments()

    # Get the metric values and aggregate
    df = eval_metrics_experiments(runs, metrics=("test", example_metric))
    agg = aggregate_metrics_feval_budget(df, metric_direction_override={"test": "-"})

    # Assert the right indices show up with the right types
    assert ptypes.is_string_dtype(agg.index.get_level_values("problem"))
    assert ptypes.is_integer_dtype(agg.index.get_level_values("exp_idx"))

    # Assert the right metrics show up
    assert ptypes.is_float_dtype(agg[("test", "mean")])
    assert ptypes.is_float_dtype(agg[("test", "std")])
    assert ptypes.is_float_dtype(agg[("test", "median")])
    assert ptypes.is_float_dtype(agg[("test", "pct_2.5")])
    assert ptypes.is_float_dtype(agg[("test", "pct_97.5")])
    assert ptypes.is_bool_dtype(agg[("test", "wilcoxon_best")])

    # Check the mean and std
    df["problem"] = df.apply(lambda x: normalize_problem_name(x["problem"]), axis=1)
    test_df = apply_feval_cutoff(df)
    for index, row in test_df.groupby(["problem", "exp_idx"]).agg({"test": ["mean", "std"]}).iterrows():
        np.testing.assert_allclose(row[("test", "mean")], agg.loc[index][("test", "mean")])


def test_aggregate_metrics_stats_test():
    """
    Test the statistical comparisons for the metric aggregation function. Several problems are generated and several different
    runs are made. The runs have evaluations with a normally distributed metric. The mean of the values is shifted to known
    positions. We then perform the comparisons and confirm the comparisons based on their centroid location.
    """
    # The centroid locations of the metric values for each problem
    run_locs = {
        "ZDT1": [0, 1, 2, 3],
        "ZDT2": [1, 2, 3, 0],
        "ZDT3": [2, 3, 0, 1],
        "ZDT4": [3, 0, 1, 2],
        "ZDT6": [1, 0, 0, 0],
        "DTLZ1": [0, 0, 0, 1],
    }

    # Construct a dataframe with metric values normally distributed so that idx
    np.random.seed(1)
    rows = []
    for exp_idx in range(4):
        for problem in run_locs:
            for run_idx, val in enumerate(np.random.normal(run_locs[problem][exp_idx], 0.1, size=64)):
                for fevals in range(10):
                    rows.append(
                        {
                            "run_idx": run_idx,
                            "exp_idx": exp_idx,
                            "problem": problem,
                            "fevals": fevals,
                            "test": val if fevals == 8 else 0.0,
                            "fname": "",
                        }
                    )
    df = pd.DataFrame(rows)

    # Aggregate to get stats test values
    agg = aggregate_metrics_feval_budget(df, max_feval=8, wilcoxon_idx=0, metric_direction_override={"test": "-"})

    # Check each problem for the comparisons
    for prob in run_locs:
        # The "normalized" name
        prob_norm = normalize_problem_name(prob)

        for exp_idx, loc in enumerate(run_locs[prob]):
            # Check the wilcoxon comparison
            if exp_idx == 0:
                wilcoxon_comp = ""
            elif loc < run_locs[prob][0]:
                wilcoxon_comp = "+"
            elif loc == run_locs[prob][0]:
                wilcoxon_comp = "="
            else:
                wilcoxon_comp = "-"
            assert agg.loc[prob_norm, exp_idx].iloc[0][("test", "wilcoxon_comp")] == wilcoxon_comp

            # Check if we are one of the best in the row
            assert agg.loc[prob_norm, exp_idx].iloc[0][("test", "wilcoxon_best")] == (loc == min(run_locs[prob]))


def test_feval_cutoff():
    """
    Test the cutoff function by making a table of values with a test metric equalling 1.0 only at fevals=7. The cutoff is made
    for fevals=7 and we check all the metric values.
    """
    # Create a test dataframe (big nasty nested loop)
    rows = []
    for run_idx in range(16):
        for exp_idx in range(4):
            for problem in ["ZDT1", "ZDT2"]:
                for fevals in range(10):
                    rows.append(
                        {
                            "run_idx": run_idx,
                            "exp_idx": exp_idx,
                            "problem": problem,
                            "fevals": fevals,
                            "test": 1.0 if fevals == 7 else 0.0,
                        }
                    )
    df = pd.DataFrame(rows)

    # Run the cutoff function
    df_cutoff = apply_feval_cutoff(df, max_feval=7)
    assert (df_cutoff["test"] == 1.0).all()


@pytest.mark.parametrize(
    "use_names, metric, problem_params",
    [
        (False, None, [("n", lambda x: x.n)]),
        (True, None, [("n", lambda x: x.n)]),
        (True, "test", [("n", lambda x: x.n)]),
        (True, None, [("n", "n")]),
        (True, None, ["n"]),
        (True, None, "n"),
    ],
)
def test_construct_metric_comparison_table(use_names, metric, problem_params):
    """
    Create a comparison table from a set of Experiment objects. Perform some basic checks that the rows and columns show up
    appropriately with and without specification of the names.
    """
    # Create some test objects
    names = ["a", "b", "c", "d"] if use_names else None
    runs = generate_moga_experiments(names=names)

    # Construct the comparison table
    df = eval_metrics_experiments(runs, metrics=("test", example_metric))
    df = aggregate_metrics_feval_budget(df, metric_direction_override={"test": "-"})
    dfm = construct_metric_comparison_table(df, metric, problem_params=problem_params)

    # Check we get the right row and column names
    indices_act = set(("ZDT1", Problem.from_line_fmt(p.problem).n) for p in runs[0].runs)
    indices_tst = set(dfm.index)
    assert indices_act == indices_tst
    assert set(dfm.columns) == set(r.name for r in runs)


def test_aggregate_metrics_series():
    """
    Test the series aggregation function by comparing with the other one (`aggregate_metrics_feval_budget`)
    """
    # Seed the RNG
    np.random.seed(0)

    # Create some test objects
    runs = generate_moga_experiments()

    # Get the metric values and aggregate
    df = eval_metrics_experiments(runs, metrics=("test", example_metric))
    agg = aggregate_metric_series(df)

    # Assert the right indices show up with the right types
    assert ptypes.is_string_dtype(agg.index.get_level_values("problem"))
    assert ptypes.is_integer_dtype(agg.index.get_level_values("exp_idx"))
    assert ptypes.is_integer_dtype(agg.index.get_level_values("fevals"))

    # Assert the right metrics show up
    assert ptypes.is_float_dtype(agg[("test", "mean")])
    assert ptypes.is_float_dtype(agg[("test", "std")])
    assert ptypes.is_float_dtype(agg[("test", "median")])
    assert ptypes.is_float_dtype(agg[("test", "pct_2.5")])
    assert ptypes.is_float_dtype(agg[("test", "pct_97.5")])

    # Check against the other aggregation method
    for exp_idx, run in enumerate(runs):
        # Get a few fevals to test the series
        test_fevals = np.random.choice(
            agg.xs(exp_idx, level="exp_idx").index.get_level_values("fevals"),
            size=3,
            replace=False,
        )

        # Test against the other aggregation function
        for fevals in test_fevals:
            for idx, actual_row in aggregate_metrics_feval_budget(
                df[df["exp_idx"] == exp_idx],
                max_feval=fevals,
                metric_direction_override={"test": "-"},
            ).iterrows():
                # Get the problem
                problem = idx[0]

                # Get the row which corresponds to this row in the test dataframe
                test_row = agg.loc[(problem, exp_idx, run.name, fevals)]

                # Compare with the "actual" value
                np.testing.assert_allclose(actual_row[("test", "mean")], test_row[("test", "mean")])


def test_comparison_table_to_latex():
    """
    Make sure that `comparison_table_to_latex` runs on reasonable input and produces something that looks like latex.
    """
    # Create some test objects
    runs = generate_moga_experiments(names=["a", "b", "c", "d"])

    # Construct the comparison table
    df = eval_metrics_experiments(runs, metrics=("test", example_metric))
    df = aggregate_metrics_feval_budget(df, metric_direction_override={"test": "-"})
    df = construct_metric_comparison_table(df, problem_params="n")

    # Test the conversion function
    latex_str = comparison_table_to_latex(df)

    # Some basic tests
    assert isinstance(latex_str, str)
    assert "tabular" in latex_str



from datetime import datetime, timezone
from functools import reduce
from pydantic import BaseModel, Field, field_validator, ConfigDict, model_validator
from typing import List, Dict, Union, Optional
import h5py
import numpy as np
import random
import re
import string

from .utils import get_domination


class Population(BaseModel):
    """
    Stores the individuals in a population for one reporting interval in a genetic algorithm. Conventional names are used for
    the decision variables (x), the objectives (f), and inequality constraints (g). The first dimension of each array is the
    batch dimension. The number of evaluations of the objective functions performed to reach this state is also recorded.
    Objectives are assumed to be part of a minimization problem and constraints are set such that individuals with g_i > 0 are
    feasible.

    All arrays must have the same size batch dimension even if they are empty. In this case the non-batch dimension will be
    zero length. Names may be associated with decision variables, objectives, or constraints in the form of lists.
    """

    x: np.ndarray
    f: np.ndarray
    g: np.ndarray
    fevals: int
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Optional lists of names for decision variables, objectives, and constraints
    names_x: Optional[List[str]] = None
    names_f: Optional[List[str]] = None
    names_g: Optional[List[str]] = None

    @model_validator(mode="before")
    @classmethod
    def set_default_vals(cls, values):
        """
        Handles automatic setting of `x`, `f`,  `g`, `fevals` when some are not specified. The arrays are set to an empty array
        with a zero length non-batch dimension. The number of function evaluations (`fevals`) is set to the number of individuals
        in the population (assuming here that each was evaluated to get to this point).
        """
        # Determine the batch size from the first non-None array
        batch_size = next(
            (arr.shape[0] for arr in [values.get("x"), values.get("f"), values.get("g")] if arr is not None),
            None,
        )
        if batch_size is None:
            raise ValueError("Must specify one of x, f, or g")

        # Set empty arrays for unspecified fields
        if values.get("x") is None:
            values["x"] = np.empty((batch_size, 0), dtype=np.float64)
        if values.get("f") is None:
            values["f"] = np.empty((batch_size, 0), dtype=np.float64)
        if values.get("g") is None:
            values["g"] = np.empty((batch_size, 0), dtype=np.float64)

        # Set fevals to number of individuals if not included
        if values.get("fevals") is None:
            values["fevals"] = batch_size
        return values

    @model_validator(mode="after")
    def validate_batch_dimensions(self):
        """
        Confirms that the arrays have the same length batch dimension.
        """
        # Validate batch dimensions
        x_size, f_size, g_size = self.x.shape[0], self.f.shape[0], self.g.shape[0]
        if len(set([x_size, f_size, g_size])) != 1:
            raise ValueError(f"Batch dimensions do not match (len(x)={x_size}, len(f)={f_size}, len(g)={g_size})")
        return self

    @model_validator(mode="after")
    def validate_names(self):
        """
        Checks that the name lists, if used, are correctly sized to the number of decision variables, objectives, or
        constraints.
        """
        if self.names_x and len(self.names_x) != self.x.shape[1]:
            raise ValueError("Length of names_x must match the number of decision variables in x.")
        if self.names_f and len(self.names_f) != self.f.shape[1]:
            raise ValueError("Length of names_f must match the number of objectives in f.")
        if self.names_g and len(self.names_g) != self.g.shape[1]:
            raise ValueError("Length of names_g must match the number of constraints in g.")
        return self

    @field_validator("x", "f", "g")
    @classmethod
    def validate_numpy_arrays(cls, value: np.ndarray, info) -> np.ndarray:
        """
        Double checks that the arrays have the right numbe of dimensions and datatype.
        """
        if value.dtype != np.float64:
            raise TypeError(f"Expected array of type { np.float64} for field '{info.field_name}', got {value.dtype}")
        if value.ndim != 2:
            raise ValueError(f"Expected array with 2 dimensions for field '{info.field_name}', got {value.ndim}")

        return value

    @field_validator("fevals")
    @classmethod
    def validate_feval(cls, v):
        if v < 0:
            raise ValueError("fevals must be a non-negative integer")
        return v

    def __eq__(self, other):
        if not isinstance(other, Population):
            return False
        return (
            np.array_equal(self.x, other.x)
            and np.array_equal(self.f, other.f)
            and np.array_equal(self.g, other.g)
            and self.fevals == other.fevals
            and self.names_x == other.names_x
            and self.names_f == other.names_f
            and self.names_g == other.names_g
        )

    def __add__(self, other: "Population") -> "Population":
        """
        The sum of two populations is defined here as the population containing all unique individuals from both (set union).
        The number of function evaluations is by default set to the sum of the function evaluations in each input population.
        """
        if not isinstance(other, Population):
            raise TypeError("Operands must be instances of Population")

        # Check that the names are consistent
        if self.names_x != other.names_x:
            raise ValueError("names_x are inconsistent between populations")
        if self.names_f != other.names_f:
            raise ValueError("names_f are inconsistent between populations")
        if self.names_g != other.names_g:
            raise ValueError("names_g are inconsistent between populations")

        # Concatenate the arrays along the batch dimension (axis=0)
        new_x = np.concatenate((self.x, other.x), axis=0)
        new_f = np.concatenate((self.f, other.f), axis=0)
        new_g = np.concatenate((self.g, other.g), axis=0)

        # Unique the arrays
        _, indices = np.unique(np.concatenate([new_x, new_f, new_g], axis=1), return_index=True, axis=0)
        new_x = new_x[indices, :]
        new_f = new_f[indices, :]
        new_g = new_g[indices, :]

        # Set fevals to the maximum of the two fevals values
        new_feval = self.fevals + other.fevals

        # Return a new Population instance
        return Population(
            x=new_x,
            f=new_f,
            g=new_g,
            fevals=new_feval,
            names_x=self.names_x,
            names_f=self.names_f,
            names_g=self.names_g,
        )

    def __getitem__(self, idx: Union[slice, np.ndarray, List[int]]) -> "Population":
        """
        Indexing operator to select along the batch dimension in the arrays.

        Parameters
        ----------
        idx : slice, np.ndarray, or list of ints
            The indices used to select along the batch dimension.

        Returns
        -------
        Population
            A new Population instance containing the selected individuals.
        """
        return Population(
            x=self.x[idx],
            f=self.f[idx],
            g=self.g[idx],
            fevals=self.fevals,
            names_x=self.names_x,
            names_f=self.names_f,
            names_g=self.names_g,
        )

    def get_nondominated_indices(self):
        """
        Returns a boolean array of whether or not an individual is non-dominated.
        """
        return np.sum(get_domination(self.f, self.g), axis=0) == 0

    def get_feasible_indices(self):
        if self.g.shape[1] == 0:
            return np.ones((len(self)), dtype=bool)
        return np.all(self.g >= 0.0, axis=1)

    def get_nondominated_set(self):
        return self[self.get_nondominated_indices()]

    @classmethod
    def from_random(
        cls,
        n_objectives: int,
        n_decision_vars: int,
        n_constraints: int,
        pop_size: int,
        fevals: int = 0,
        generate_names: bool = False,
    ) -> "Population":
        """
        Generate a randomized instance of the Population class.

        Parameters
        ----------
        n_objectives : int
            The number of objectives for each individual.
        n_decision_vars : int
            The number of decision variables for each individual.
        n_constraints : int
            The number of inequality constraints for each individual.
        pop_size : int
            The number of individuals in the population.
        fevals : int, optional
            The number of evaluations of the objective functions performed to reach this state, by default 0.
        generate_names : bool, optional
            Whether to include names for the decision variables, objectives, and constraints, by default False.

        Returns
        -------
        Population
            An instance of the Population class with random values for decision variables (`x`), objectives (`f`),
            and inequality constraints (`g`). Optionally, names for these components can be included.

        Examples
        --------
        >>> random_population = Population.from_random(n_objectives=3, n_decision_vars=5, n_constraints=2, pop_size=10)
        >>> print(random_population.x.shape)
        (10, 5)
        >>> print(random_population.f.shape)
        (10, 3)
        >>> print(random_population.g.shape)
        (10, 2)
        >>> print(random_population.fevals)
        0
        """
        x = np.random.rand(pop_size, n_decision_vars)
        f = np.random.rand(pop_size, n_objectives)
        g = np.random.rand(pop_size, n_constraints) if n_constraints > 0 else np.empty((pop_size, 0))

        # Optionally generate names if generate_names is True
        names_x = [f"x{i+1}" for i in range(n_decision_vars)] if generate_names else None
        names_f = [f"f{i+1}" for i in range(n_objectives)] if generate_names else None
        names_g = [f"g{i+1}" for i in range(n_constraints)] if generate_names else None

        return cls(
            x=x,
            f=f,
            g=g,
            fevals=fevals,
            names_x=names_x,
            names_f=names_f,
            names_g=names_g,
        )

    def __len__(self):
        return self.x.shape[0]

    def __repr__(self) -> str:
        return f"Population(size={len(self)}, vars={self.x.shape[1]}, objs={self.f.shape[1]}, cons={self.g.shape[1]}, fevals={self.fevals})"

    def __str__(self):
        return self.__repr__()

    def count_unique_individuals(self, decimals=13):
        """
        Calculates the number of unique individuals in the population. Uses `np.round` to avoid floating point accuracy issues.

        Parameters
        ----------
        decimals : int, optional
            _description_, by default 13

        Returns
        -------
        int
            The number of unique individuals
        """
        features = np.concatenate((self.x, self.f, self.g), axis=1)
        return np.unique(features.round(decimals=decimals), axis=0).shape[0]

    @property
    def n(self):
        """
        The number of decision variables.
        """
        return self.x.shape[1]

    @property
    def m(self):
        """
        The number of objectives.
        """
        return self.f.shape[1]

    @property
    def n_constraints(self):
        """
        The number of constraints.
        """
        return self.g.shape[1]


class History(BaseModel):
    """
    Contains populations output from a multiobjective genetic algorithm at some reporting interval in order to track
    its history as it converges to a solution.

    Assumptions:
     - All reports must have a consistent number of objectives, decision variables, and constraints.
     - Names, if used, must be consistent across populations
    """

    reports: List[Population]
    problem: str
    metadata: Dict[str, Union[str, int, float, bool]] = Field(default_factory=dict)

    @model_validator(mode="after")
    def validate_consistent_populations(self):
        """
        Makes sure that all populations have the same number of objectives, constraints, and decision variables.
        """
        n_decision_vars = [x.x.shape[1] for x in self.reports]
        n_objectives = [x.f.shape[1] for x in self.reports]
        n_constraints = [x.g.shape[1] for x in self.reports]

        if n_decision_vars and len(set(n_decision_vars)) != 1:
            raise ValueError(f"Inconsistent number of decision variables in reports: {n_decision_vars}")
        if n_objectives and len(set(n_objectives)) != 1:
            raise ValueError(f"Inconsistent number of objectives in reports: {n_objectives}")
        if n_constraints and len(set(n_constraints)) != 1:
            raise ValueError(f"Inconsistent number of constraints in reports: {n_constraints}")

        # Validate consistency of names across populations
        names_x = [tuple(x.names_x) if x.names_x is not None else None for x in self.reports]
        names_f = [tuple(x.names_f) if x.names_f is not None else None for x in self.reports]
        names_g = [tuple(x.names_g) if x.names_g is not None else None for x in self.reports]

        # If names are provided, check consistency
        if names_x and len(set(names_x)) != 1:
            raise ValueError(f"Inconsistent names for decision variables in reports: {names_x}")
        if names_f and len(set(names_f)) != 1:
            raise ValueError(f"Inconsistent names for objectives in reports: {names_f}")
        if names_g and len(set(names_g)) != 1:
            raise ValueError(f"Inconsistent names for constraints in reports: {names_g}")

        return self

    def __eq__(self, other):
        if not isinstance(other, History):
            return False
        return self.reports == other.reports and self.problem == other.problem and self.metadata == other.metadata

    @classmethod
    def from_random(
        cls,
        n_populations: int,
        n_objectives: int,
        n_decision_vars: int,
        n_constraints: int,
        pop_size: int,
        generate_names: bool = False,
    ) -> "History":
        """
        Generate a randomized instance of the History class, including random problem name and metadata.

        Parameters
        ----------
        n_populations : int
            The number of populations (reports) to generate.
        n_objectives : int
            The number of objectives for each individual in each population.
        n_decision_vars : int
            The number of decision variables for each individual in each population.
        n_constraints : int
            The number of inequality constraints for each individual in each population.
        pop_size : int
            The number of individuals in each population.
        generate_names : bool, optional
            Whether to include names for the decision variables, objectives, and constraints, by default False.

        Returns
        -------
        History
            An instance of the History class with a list of randomly generated Population instances, a random problem name, and
            random metadata.
        """
        # Randomly generate a problem name
        problem = f"Optimization_Problem_{random.randint(1000, 9999)}"

        # Randomly generate metadata
        metadata_keys = ["author", "version", "description", "date"]
        metadata_values = [
            "".join(random.choices(string.ascii_uppercase + string.digits, k=5)),
            random.uniform(1.0, 3.0),
            "Randomly generated metadata",
            f"{random.randint(2020, 2024)}-0{random.randint(1, 9)}-{random.randint(10, 28)}",
        ]
        metadata = dict(zip(metadata_keys, metadata_values))

        # Generate populations with increasing fevals values
        reports = [
            Population.from_random(
                n_objectives,
                n_decision_vars,
                n_constraints,
                pop_size,
                fevals=(i + 1) * pop_size,
                generate_names=generate_names,
            )
            for i in range(n_populations)
        ]

        return cls(reports=reports, problem=problem, metadata=metadata)

    def _to_h5py_group(self, g: h5py.Group):
        """
        Store the history object in an HDF5 group. The populations are concatenated together and stored together as a few small
        arrays rather than their own groups to limit the number of python calls to the HDF5 API. This is done for performance
        reasons and the speed of both variants (concatenated ararys and populations in groups) were benchmarked with the
        concatenated arrays performing up to 10x faster on combined write/read tests.

        Parameters
        ----------
        g : h5py.Group
            The h5py group to write our data to.
        """
        # Save the metadata
        g.attrs["problem"] = self.problem
        g_md = g.create_group("metadata")
        for k, v in self.metadata.items():
            g_md.attrs[k] = v

        # Save data from each population into one bigger dataset to reduce API calls to HDF5 file reader
        g.attrs["pop_sizes"] = [len(r) for r in self.reports]
        g.attrs["fevals"] = [r.fevals for r in self.reports]
        if self.reports:
            g["x"] = np.concatenate([r.x for r in self.reports], axis=0)
            g["f"] = np.concatenate([r.f for r in self.reports], axis=0)
            g["g"] = np.concatenate([r.g for r in self.reports], axis=0)
        else:
            g["x"] = np.empty(())
            g["f"] = np.empty(())
            g["g"] = np.empty(())

        # Save names
        if self.reports and self.reports[0].names_x is not None:
            g["x"].attrs["names"] = self.reports[0].names_x
        if self.reports and self.reports[0].names_f is not None:
            g["f"].attrs["names"] = self.reports[0].names_f
        if self.reports and self.reports[0].names_g is not None:
            g["g"].attrs["names"] = self.reports[0].names_g

    @classmethod
    def _from_h5py_group(cls, grp: h5py.Group):
        """
        Construct a new History object from data in an HDF5 group.

        Parameters
        ----------
        grp : h5py.Group
            The group containing the history data.

        Returns
        -------
        History
            The loaded history object
        """
        # Get the decision vars, objectives, and constraints
        x = grp["x"][()]
        f = grp["f"][()]
        g = grp["g"][()]

        # Get the names
        names_x = grp["x"].attrs.get("names", None)
        names_f = grp["f"].attrs.get("names", None)
        names_g = grp["g"].attrs.get("names", None)

        # Create the population objects
        start_idx = 0
        reports = []
        for pop_size, fevals in zip(grp.attrs["pop_sizes"], grp.attrs["fevals"]):
            reports.append(
                Population(
                    x=x[start_idx : start_idx + pop_size],
                    f=f[start_idx : start_idx + pop_size],
                    g=g[start_idx : start_idx + pop_size],
                    fevals=fevals,
                    names_x=names_x,
                    names_f=names_f,
                    names_g=names_g,
                )
            )
            start_idx += pop_size

        # Return as a history object
        return cls(
            problem=grp.attrs["problem"],
            reports=reports,
            metadata={k: v for k, v in grp["metadata"].attrs.items()},
        )

    def to_nondominated(self):
        """
        Returns a history object with the same number of population objects, but the individuals in each generation are the
        nondominated solutions seen up to this point. The function evaluation count is unchanged.

        Returns
        -------
        History
            History object containing the nondominated solution
        """
        if len(self.reports) < 2:
            return self

        def pf_reduce(a, b):
            return a + [(a[-1] + b).get_nondominated_set()]

        # Get the nondominated objectives
        new_reports = reduce(pf_reduce, self.reports[1:], [self.reports[0].get_nondominated_set()])

        # Make sure fevals carries over
        for n, o in zip(new_reports, self.reports):
            n.fevals = o.fevals

        return History(reports=new_reports, problem=self.problem, metadata=self.metadata.copy())

    def __repr__(self) -> str:
        dims = (
            (
                f"vars={self.reports[0].x.shape[1]}, objs={self.reports[0].f.shape[1]}, "
                f"cons={self.reports[0].g.shape[1]}"
            )
            if self.reports
            else "empty"
        )
        return f"History(problem='{self.problem}', reports={len(self.reports)}, {dims})"

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self.reports)


class Experiment(BaseModel):
    """
    Represents on "experiment" performed on a multibojective genetic algorithm. It may contain several evaluations of the
    algorithm on different problems or repeated iterations on the same problem.
    """

    runs: List[History]
    name: str
    author: str = ""
    software: str = ""
    software_version: str = ""
    comment: str = ""
    creation_time: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    file_version: str = "1.0.0"

    def __eq__(self, other):
        if not isinstance(other, Experiment):
            return False
        return (
            self.name == other.name
            and self.author == other.author
            and self.software == other.software
            and self.software_version == other.software_version
            and self.comment == other.comment
            and self.creation_time == other.creation_time
            and self.runs == other.runs
        )

    def __repr__(self) -> str:
        metadata = [
            f"name='{self.name}'",
            f"created='{self.creation_time.strftime('%Y-%m-%d')}'",
        ]
        if self.author:
            metadata.append(f"author='{self.author}'")
        if self.software:
            version = f" {self.software_version}" if self.software_version else ""
            metadata.append(f"software='{self.software}{version}'")
        metadata.append(f"runs={len(self.runs)}")
        return f"Experiment({', '.join(metadata)})"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_random(
        cls,
        n_histories: int,
        n_populations: int,
        n_objectives: int,
        n_decision_vars: int,
        n_constraints: int,
        pop_size: int,
        generate_names: bool = False,
    ) -> "Experiment":
        """
        Generate a randomized instance of the Experiment class.

        Parameters
        ----------
        n_histories : int
            The number of histories to generate.
        n_populations : int
            The number of populations in each history.
        n_objectives : int
            The number of objectives for each individual in each population.
        n_decision_vars : int
            The number of decision variables for each individual in each population.
        n_constraints : int
            The number of inequality constraints for each individual in each population.
        pop_size : int
            The number of individuals in each population.
        generate_names : bool, optional
            Whether to include names for the decision variables, objectives, and constraints, by default False.

        Returns
        -------
        Experiment
            An instance of the Experiment class with a list of randomly generated History instances and a random name.
        """
        # Generate random histories
        runs = [
            History.from_random(
                n_populations,
                n_objectives,
                n_decision_vars,
                n_constraints,
                pop_size,
                generate_names=generate_names,
            )
            for _ in range(n_histories)
        ]

        # Randomly generate an name for the experiment
        name = f"Experiment_{random.randint(1000, 9999)}"

        # Generate random values or placeholders for other attributes
        author = f"Author_{random.randint(1, 100)}"
        software = f"Software_{random.randint(1, 10)}"
        software_version = f"{random.randint(1, 5)}.{random.randint(0, 9)}"
        comment = "Randomly generated experiment"

        return cls(
            runs=runs,
            name=name,
            author=author,
            software=software,
            software_version=software_version,
            comment=comment,
        )

    def save(self, fname):
        """
        Saves the experiment data into an HDF5 file at the specified filename.

        Parameters
        ----------
        fname : str
            Filename to save to
        """
        with h5py.File(fname, mode="w") as f:
            # Save metadata as attributes
            f.attrs["name"] = self.name
            f.attrs["author"] = self.author
            f.attrs["software"] = self.software
            f.attrs["software_version"] = self.software_version
            f.attrs["comment"] = self.comment
            f.attrs["creation_time"] = self.creation_time.isoformat()
            f.attrs["file_version"] = self.file_version
            f.attrs["file_format"] = "ParetoBench Multi-Objective Optimization Data"

            # Calculate the necessary zero padding based on the number of runs
            max_len = len(str(len(self.runs) - 1))

            # Save each run into its own group
            for idx, run in enumerate(self.runs):
                run._to_h5py_group(f.create_group(f"run_{idx:0{max_len}d}"))

    @classmethod
    def load(cls, fname):
        """
        Creates a new Experiment object from an HDF5 file on disk.

        Parameters
        ----------
        fname : str
            Filename to load the data from

        Returns
        -------
        Experiment
            The loaded experiment object

        Examples
        --------
        >>> exp = Experiment.load('state_of_the_art_algorithm_benchmarking_data.h5')
        >>> print(exp.name)
        NewAlg
        >>> print(len(exp.runs))
        64
        """
        # Load the data
        with h5py.File(fname, mode="r") as f:
            # Load each of the runs keeping track of the order of the indices
            idx_runs = []
            for idx_str, run_grp in f.items():
                m = re.match(r"run_(\d+)", idx_str)
                if m:
                    idx_runs.append((int(m.group(1)), History._from_h5py_group(run_grp)))
            runs = [x[1] for x in sorted(idx_runs, key=lambda x: x[0])]

            # Convert the creation_time back to a timezone-aware datetime object
            creation_time = datetime.fromisoformat(f.attrs["creation_time"]).astimezone(timezone.utc)

            # Return as an experiment object
            return cls(
                runs=runs,
                name=f.attrs["name"],
                author=f.attrs["author"],
                software=f.attrs["software"],
                software_version=f.attrs["software_version"],
                comment=f.attrs["comment"],
                creation_time=creation_time,
            )



import numpy as np

from .problem import Problem, ProblemWithPF
from .utils import weighted_chunk_sizes
from .containers import Population


class ZDTx(Problem, ProblemWithPF):
    n: int = 30

    @property
    def m(self):
        return 2

    @property
    def reference(self):
        return (
            "Zitzler, E., Deb, K., & Thiele, L. (2000). Comparison of Multiobjective Evolutionary Algorithms: Empirical Results. "
            "Evolutionary Computation, 8(2), 173–195. https://doi.org/10.1162/106365600568202"
        )

    @property
    def var_lower_bounds(self):
        return np.zeros(self.n)

    @property
    def var_upper_bounds(self):
        return np.ones(self.n)


class ZDT1(ZDTx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        g = 1 + 9 * np.sum(x[1:], axis=0) / (self.n - 1)
        return Population(
            f=np.array(
                [
                    x[0],
                    g * (1 - np.sqrt(x[0] / g)),
                ]
            ).T
        )

    def get_pareto_front(self, n):
        x = np.linspace(0, 1, n)
        return np.array([x, 1 - np.sqrt(x)]).T


class ZDT2(ZDTx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        g = 1 + 9 * np.sum(x[1:], axis=0) / (self.n - 1)
        return Population(
            f=np.array(
                [
                    x[0],
                    g * (1 - (x[0] / g) ** 2),
                ]
            ).T
        )

    def get_pareto_front(self, n):
        x = np.linspace(0, 1, n)
        return np.array([x, 1 - x**2]).T


class ZDT3(ZDTx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        g = 1 + 9 * np.sum(x[1:], axis=0) / (self.n - 1)
        return Population(
            f=np.array(
                [
                    x[0],
                    g * (1 - np.sqrt(x[0] / g) - x[0] / g * np.sin(10 * np.pi * x[0])),
                ]
            ).T
        )

    def get_pareto_front(self, n):
        # The non-dominated regions from dev_notebooks\zdt3_pareto_front.ipynb
        regs = [
            (0.0, 0.08300152991598536),
            (0.18222872351573624, 0.25776236490626825),
            (0.4093136698086569, 0.453882099086383),
            (0.6183967894392438, 0.6525116988130034),
            (0.8233317933269128, 0.851832862382744),
        ]

        # Evaluate along the non-dominated region
        f = []
        for r, my_n in zip(regs, weighted_chunk_sizes(n, [r[1] - r[0] for r in regs])):
            x = np.linspace(*r, my_n)
            f.append([x, 1 - np.sqrt(x) - x * np.sin(10 * np.pi * x)])
        return np.concatenate(f, axis=1).T


class ZDT4(ZDTx):
    n: int = 10

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        g = 1 + 10 * (self.n - 1) + np.sum(x[1:] ** 2 - 10 * np.cos(4 * np.pi * x[1:]), axis=0)
        return Population(
            f=np.array(
                [
                    x[0],
                    g * (1 - np.sqrt(x[0] / g)),
                ]
            ).T
        )

    @property
    def var_lower_bounds(self):
        return np.concatenate(([0], -5 * np.ones(self.n - 1)))

    @property
    def var_upper_bounds(self):
        return np.concatenate(([1], 5 * np.ones(self.n - 1)))

    def get_pareto_front(self, n):
        x = np.linspace(0, 1, n)
        return np.array([x, 1 - np.sqrt(x)]).T


class ZDT6(ZDTx):
    n: int = 10

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        f1 = 1 - np.exp(-4 * x[0]) * (np.sin(6 * np.pi * x[0])) ** 6
        g = 1 + 9 * (np.sum(x[1:], axis=0) / (self.n - 1)) ** 0.25
        return Population(
            f=np.array(
                [
                    f1,
                    g * (1 - (f1 / g) ** 2),
                ]
            ).T
        )

    def get_pareto_front(self, n):
        # From dev_notebooks\zdt6_pareto_front.ipynb
        x = np.linspace(0.28077531881538886, 1, n)
        return np.array([x, 1 - np.power(x, 2)]).T



from matplotlib import animation
from typing import Optional, Tuple, Literal, Union, List
import matplotlib.pyplot as plt
import numpy as np

from ..containers import History
from ..exceptions import EmptyPopulationError, NoObjectivesError
from .attainment import get_reference_point
from .population import (
    population_obj_scatter,
    population_dvar_pairs,
)
from .utils import selection_to_indices


def history_obj_scatter(
    history,
    reports: Optional[Union[int, slice, List[int], Tuple[int, int]]] = None,
    fig=None,
    ax=None,
    domination_filt: Literal["all", "dominated", "non-dominated"] = "all",
    feasibility_filt: Literal["all", "feasible", "infeasible"] = "all",
    show_points: bool = True,
    n_pf: int = 1000,
    pf_objectives: Optional[np.ndarray] = None,
    show_attainment: bool = False,
    show_dominated_area: bool = False,
    ref_point: Optional[Tuple[float, float]] = None,
    ref_point_padding: float = 0.05,
    legend_loc: Optional[str] = None,
    scale: Optional[np.ndarray] = None,
    flip_objs: bool = False,
    show_names: bool = True,
    show_pf: bool = False,
    colormap: str = "viridis",
    cmap_label: Optional[str] = None,
    generation_mode: Literal["cmap", "cumulative"] = "cmap",
    single_color: Optional[str] = None,
    label_mode: Literal["index", "fevals"] = "index",
):
    """
    Plot the objectives from a history of populations, using either a colormap for generations
    or merging all generations into a single population.

    Parameters
    ----------
    history : History object
        The history containing populations to plot
    reports : int, slice, List[int], or Tuple[int, int], optional
        Specifies which generations to plot. See `selection_to_indices` for more details.
    fig : matplotlib figure, optional
        Figure to plot on, by default None
    ax : matplotlib axis, optional
        Axis to plot on, by default None
    domination_filt : Literal["all", "dominated", "non-dominated"], optional
        Plot only the dominated/non-dominated solutions, or all. Defaults to all
    feasibility_filt : Literal['all', 'feasible', 'infeasible'], optional
        Plot only the feasible/infeasible solutions, or all. Defaults to all
    show_points : bool
        Whether to actually show the points (useful for only showing attainment surface or dominated region)
    n_pf : int, optional
        The number of points used for plotting the Pareto front (when problem allows user selectable number of points)
    pf_objectives : array-like, optional
        User-specified Pareto front objectives. Should be a 2D array where each row represents a point
        on the Pareto front and each column represents an objective value.
    show_attainment : bool, optional
        Whether to plot the attainment surface, by default False
    show_dominated_area : bool, optional
        Plots the dominated region towards the larger values of each decision var
    ref_point : Union[str, Tuple[float, float]], optional
        Where to stop plotting the dominated region / attainment surface. Must be a point to the upper right (increasing
        value of objectives in 3D) of all plotted points. By default, will set to right of max of each objective plus
        padding.
    ref_point_padding : float
        Amount of padding to apply to the automatic reference point calculation.
    legend_loc : str, optional
        Passed to `loc` argument of plt.legend
    scale : array-like, optional
        Scale factors for each objective. Must have the same length as the number of objectives.
        If None, no scaling is applied.
    flip_objs : bool, optional
        Flips the order the objectives are plotted in. IE swaps axes in 2D, reverse them in 3D.
    show_names : bool, optional
        Whether to show the names of the objectives if provided by population
    show_pf : bool, optional
        Whether to plot the Pareto front, by default True
    colormap : str, optional
        Name of the colormap to use for generation colors, by default 'viridis'
    cmap_label: Optional[str] = "Generation"
        Label for colorbar (only used when generation_mode is 'cmap')
    generation_mode: Literal['cmap', 'cumulative'] = 'cmap'
        How to handle multiple generations:
        'cmap': Plot each generation separately with colors from colormap
        'cumulative': Merge all selected generations into single population
    single_color: Optional[str] = None
        Color to use when generation_mode is 'cumulative'. If None, uses default color from matplotlib.
    label_mode: Literal['index', 'fevals'] = 'index'
        Whether to use report index or function evaluations (fevals) for labels

    Returns
    -------
    matplotlib figure and matplotlib axis
        The figure and axis containing the history plot
    """
    # Basic input validation
    if not history.reports:
        raise ValueError("History contains no reports")

    # Handle different types of selection
    indices = selection_to_indices(reports, len(history.reports))

    if not indices:
        raise ValueError("No reports selected")

    # Get dimensions from first population
    first_pop = history.reports[0]
    if not len(first_pop):
        raise EmptyPopulationError()
    if not first_pop.m:
        raise NoObjectivesError()

    if fig is None:
        fig = plt.figure()

    # Create base settings for population_obj_scatter
    obj_settings = dict(
        domination_filt=domination_filt,
        feasibility_filt=feasibility_filt,
        show_points=show_points,
        n_pf=n_pf,
        show_names=show_names,
        show_attainment=show_attainment,
        show_dominated_area=show_dominated_area,
        pf_objectives=pf_objectives,
        legend_loc=legend_loc,
        scale=scale,
        flip_objs=flip_objs,
    )

    # Calculate global reference point if not provided
    if ref_point is None:
        combined_population = history.reports[indices[0]]
        for idx in indices[1:]:
            combined_population = combined_population + history.reports[idx]
        obj_settings["ref_point"] = get_reference_point(combined_population, padding=ref_point_padding)
    else:
        obj_settings["ref_point"] = ref_point

    if generation_mode == "cumulative":
        # Merge all selected populations
        combined_population = history.reports[indices[0]]
        for idx in indices[1:]:
            combined_population = combined_population + history.reports[idx]

        # Set optional color and plot combined population
        obj_settings["color"] = single_color  # Will use default if None
        if show_pf and pf_objectives is not None:
            obj_settings["pf_objectives"] = pf_objectives
        elif show_pf and history.problem is not None:
            obj_settings["problem"] = history.problem

        fig, ax = population_obj_scatter(combined_population, fig=fig, ax=ax, **obj_settings)

    elif generation_mode == "cmap":
        if label_mode == "index":
            norm_values = indices
            label = cmap_label if cmap_label else "Generation"
        else:  # fevals mode
            norm_values = [history.reports[idx].fevals for idx in indices]
            label = cmap_label if cmap_label else "Function Evaluations"

        cmap = plt.get_cmap(colormap)
        norm = plt.Normalize(min(norm_values), max(norm_values))

        # Plot each selected generation
        for plot_idx, gen_idx in enumerate(indices):
            population = history.reports[gen_idx]
            norm_value = gen_idx if label_mode == "index" else population.fevals
            obj_settings["color"] = cmap(norm(norm_value))
            obj_settings["dominated_area_zorder"] = -2 - plot_idx

            # Only plot PF on the last iteration if requested
            if plot_idx == len(indices) - 1 and show_pf and pf_objectives is not None:
                obj_settings["pf_objectives"] = pf_objectives
            elif plot_idx == len(indices) - 1 and show_pf and history.problem is not None:
                obj_settings["problem"] = history.problem
            else:
                obj_settings["pf_objectives"] = None
                obj_settings["problem"] = None

            # Plot this generation
            fig, ax = population_obj_scatter(population, fig=fig, ax=ax, **obj_settings)

        # Add colorbar if label is provided
        if label:
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            sm.set_array([])  # Dummy array for the colorbar
            fig.colorbar(sm, ax=ax, label=label)
    else:
        raise ValueError(f"Unrecognized generation mode: {generation_mode}")

    return fig, ax


def history_dvar_pairs(
    history: History,
    reports: Optional[Union[int, slice, List[int], tuple[int, int]]] = None,
    dvars: Optional[Union[int, slice, List[int], Tuple[int, int]]] = None,
    fig=None,
    axes=None,
    domination_filt: Literal["all", "dominated", "non-dominated"] = "all",
    feasibility_filt: Literal["all", "feasible", "infeasible"] = "all",
    hist_bins: Optional[int] = None,
    show_names: bool = True,
    lower_bounds: Optional[np.ndarray] = None,
    upper_bounds: Optional[np.ndarray] = None,
    scale: Optional[np.ndarray] = None,
    colormap: str = "viridis",
    cmap_label: Optional[str] = None,
    generation_mode: Literal["cmap", "cumulative"] = "cmap",
    single_color: Optional[str] = None,
    plot_bounds: bool = False,
    label_mode: Literal["index", "fevals"] = "index",
):
    """
    Plot the decision variables from a history of populations, using either a colormap
    for generations or merging all generations into a single population.

    Parameters
    ----------
    history : History object
        The history containing populations to plot
    reports : int, slice, List[int], or Tuple[int, int], optional
        Specifies which generations to plot. See `selection_to_indices` for more details.
    dvars : int, slice, List[int], or Tuple[int, int], optional
        Which decision vars to plot. See `population_dvar_pairs` docstring for more details.
    fig : matplotlib figure, optional
        Figure to plot on, by default None
    axes : array of matplotlib axes, optional
        Axes to plot on, by default None
    domination_filt : Literal["all", "dominated", "non-dominated"], optional
        Plot only the dominated/non-dominated solutions, or all. Defaults to all
    feasibility_filt : Literal['all', 'feasible', 'infeasible'], optional
        Plot only the feasible/infeasible solutions, or all. Defaults to all
    hist_bins : int, optional
        Number of bins for histograms on the diagonal, default is to let matplotlib choose
    show_names : bool, optional
        Whether to include variable names on the axes if they exist, default is True
    lower_bounds : array-like, optional
        Lower bounds for each decision variable
    upper_bounds : array-like, optional
        Upper bounds for each decision variable
    scale : array-like, optional
        Scale factors for each variable. Must have the same length as the number of decision vars.
        If None, no scaling is applied.
    colormap : str, optional
        Name of the colormap to use for generation colors, by default 'viridis'
    cmap_label: Optional[str] = "Generation"
        Label for colorbar (only used when generation_mode is 'cmap')
    generation_mode: Literal['cmap', 'cumulative'] = 'cmap'
        How to handle multiple generations:
        'cmap': Plot each generation separately with colors from colormap
        'cumulative': Merge all selected generations into single population
    single_color: Optional[str] = None
        Color to use when generation_mode is 'cumulative'. If None, uses default color from matplotlib.
    plot_bounds: bool = False
        Whether to plot bounds for the problem
    label_mode: Literal['index', 'fevals'] = 'index'
        Whether to use report index or function evaluations (fevals) for labels

    Returns
    -------
    matplotlib figure and array of matplotlib axes
        The figure and axes containing the history plot
    """
    # Basic input validation
    if not history.reports:
        raise ValueError("History contains no reports")

    # Handle different types of selection
    indices = selection_to_indices(reports, len(history.reports))

    if not indices:
        raise ValueError("No generations selected")

    # Get dimensions from first population
    first_pop = history.reports[0]
    if not len(first_pop):
        raise EmptyPopulationError()

    # Check if the user gave us bounds to use
    user_specified_bounds = (lower_bounds is not None) or (upper_bounds is not None)

    # Create base settings for population_dvar_pairs
    plot_settings = dict(
        domination_filt=domination_filt,
        feasibility_filt=feasibility_filt,
        hist_bins=hist_bins,
        show_names=show_names,
        scale=scale,
    )

    if generation_mode == "cumulative":
        # Merge all selected populations
        combined_population = history.reports[indices[0]]
        for idx in indices[1:]:
            combined_population = combined_population + history.reports[idx]

        # Deal with passing along the bounds
        if plot_bounds and user_specified_bounds:
            plot_settings["lower_bounds"] = lower_bounds
            plot_settings["upper_bounds"] = upper_bounds
        elif plot_bounds and history.problem is not None:
            plot_settings["problem"] = history.problem

        # Set optional color and plot combined population
        plot_settings["color"] = single_color  # Will use default if None
        fig, axes = population_dvar_pairs(combined_population, dvars=dvars, fig=fig, axes=axes, **plot_settings)

    elif generation_mode == "cmap":
        if label_mode == "index":
            norm_values = indices
            label = cmap_label if cmap_label else "Generation"
        else:  # fevals mode
            norm_values = [history.reports[idx].fevals for idx in indices]
            label = cmap_label if cmap_label else "Function Evaluations"

        cmap = plt.get_cmap(colormap)
        norm = plt.Normalize(min(norm_values), max(norm_values))

        # Plot each selected generation
        for plot_idx, gen_idx in enumerate(indices):
            population = history.reports[gen_idx]
            norm_value = gen_idx if label_mode == "index" else population.fevals
            plot_settings["color"] = cmap(norm(norm_value))

            # Only plot bounds on the last iteration if requested
            if plot_idx == len(indices) - 1:
                if plot_bounds and user_specified_bounds:
                    plot_settings["lower_bounds"] = lower_bounds
                    plot_settings["upper_bounds"] = upper_bounds
                elif plot_bounds and history.problem is not None:
                    plot_settings["problem"] = history.problem
            else:
                plot_settings["problem"] = None
                plot_settings["lower_bounds"] = None
                plot_settings["upper_bounds"] = None

            # Plot this generation
            fig, axes = population_dvar_pairs(population, dvars=dvars, fig=fig, axes=axes, **plot_settings)

        # Add colorbar if label is provided
        if label:
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            sm.set_array([])  # Dummy array for the colorbar
            # Use ax parameter to let constrained_layout handle the positioning
            fig.colorbar(sm, ax=axes.ravel().tolist(), label=label)
    else:
        raise ValueError(f"Unrecognized generation mode: {generation_mode}")

    return fig, axes


def history_obj_animation(
    history: History,
    reports: Optional[Union[int, slice, List[int], Tuple[int, int]]] = None,
    interval: int = 200,
    domination_filt: Literal["all", "dominated", "non-dominated"] = "all",
    feasibility_filt: Literal["all", "feasible", "infeasible"] = "all",
    show_points: bool = True,
    n_pf: int = 1000,
    pf_objectives: Optional[np.ndarray] = None,
    show_attainment: bool = False,
    show_dominated_area: bool = False,
    ref_point: Optional[Tuple[float, float]] = None,
    ref_point_padding: float = 0.05,
    legend_loc: Optional[str] = "upper right",
    scale: Optional[np.ndarray] = None,
    flip_objs: bool = False,
    show_names: bool = True,
    show_pf: bool = False,
    single_color: Optional[str] = None,
    dynamic_scaling: bool = False,
    cumulative: bool = False,
    scale_padding: float = 0.05,
) -> animation.Animation:
    """
    Creates an animated visualization of how the Pareto front evolves across generations.

    Parameters
    ----------
    history : paretobench History
        The history object containing populations with data to plot
    reports : int, slice, List[int], or Tuple[int, int], optional
        Specifies which generations to animate. See `selection_to_indices` for more details.
    interval : int, optional
        Delay between frames in milliseconds, by default 200
    domination_filt : Literal["all", "dominated", "non-dominated"], optional
        Plot only the dominated/non-dominated solutions, or all. Defaults to all
    feasibility_filt : Literal['all', 'feasible', 'infeasible'], optional
        Plot only the feasible/infeasible solutions, or all. Defaults to all
    show_points : bool
        Whether to actually show the points (useful for only showing attainment surface or dominated region)
    n_pf : int, optional
        The number of points used for plotting the Pareto front (when problem allows user selectable number of points)
    pf_objectives : array-like, optional
        User-specified Pareto front objectives. Should be a 2D array where each row represents a point
        on the Pareto front and each column represents an objective value.
    show_attainment : bool, optional
        Whether to plot the attainment surface, by default False
    show_dominated_area : bool, optional
        Plots the dominated region towards the larger values of each decision var
    ref_point : Union[str, Tuple[float, float]], optional
        Where to stop plotting the dominated region / attainment surface. Must be a point to the upper right (increasing
        value of objectives in 3D) of all plotted points. By default, will set to right of max of each objective plus
        padding.
    ref_point_padding : float
        Amount of padding to apply to the automatic reference point calculation.
    legend_loc : str, optional
        Passed to `loc` argument of plt.legend
    scale : array-like, optional
        Scale factors for each objective. Must have the same length as the number of objectives.
        If None, no scaling is applied.
    flip_objs : bool, optional
        Flips the order the objectives are plotted in. IE swaps axes in 2D, reverse them in 3D.
    show_names : bool, optional
        Whether to show the names of the objectives if provided by population
    show_pf : bool, optional
        Whether to plot the Pareto front, by default True
    single_color: Optional[str] = None
        Color to use when generation_mode is 'cumulative'. If None, uses default color from matplotlib.
    dynamic_scaling : bool, optional
        If True, axes limits will update based on each frame's data.
        If False, axes limits will be fixed based on all data, by default False
    cumulative : bool, optional
        If True, shows all points seen up to current frame.
        If False, shows only current frame's points, by default False
    scale_padding : float, optional
        Padding used when calculating axis limits w/ dynamic limits

    Returns
    -------
    animation.Animation
        The animation object that can be displayed in notebooks or saved to file
    """
    if not history.reports:
        raise ValueError("No populations in history to animate")

    if scale is not None:
        scale = np.asarray(scale)
        if len(scale.shape) != 1 or scale.shape[0] != history.reports[0].m:
            raise ValueError(
                f"Length of scale must match number of objectives. Got scale factors with shape {scale.shape}"
                f" and {history.reports[0].f.shape[1]} objectives."
            )
    else:
        scale = np.ones(history.reports[0].m)

    # Handle different types of selection
    indices = selection_to_indices(reports, len(history.reports))

    if not indices:
        raise ValueError("No reports selected")

    # Get dimensions from first population
    n_objectives = history.reports[indices[0]].m
    if n_objectives > 3:
        raise ValueError(f"Cannot animate more than three objectives: n_objs={n_objectives}")

    # Set up the figure based on dimensionality
    fig = plt.figure()
    if n_objectives == 3:
        ax = fig.add_subplot(111, projection="3d")
    else:
        ax = fig.add_subplot(111)

    # Calculate global axis limits if not using dynamic scaling
    if not dynamic_scaling and n_objectives == 2:
        selected_reports = [history.reports[idx] for idx in indices]
        all_f = np.vstack([pop.f for pop in selected_reports])
        xlim = (np.min(all_f[:, 0]), np.max(all_f[:, 0]))
        ylim = (np.min(all_f[:, 1]), np.max(all_f[:, 1]))
        xlim = (xlim[0] - (xlim[1] - xlim[0]) * scale_padding, xlim[1] + (xlim[1] - xlim[0]) * scale_padding)
        ylim = (ylim[0] - (ylim[1] - ylim[0]) * scale_padding, ylim[1] + (ylim[1] - ylim[0]) * scale_padding)

    # Function to update frame for animation
    def update(frame_idx):
        ax.clear()

        # Get the actual generation index from our selected indices
        gen_idx = indices[frame_idx]

        # Calculate the slice for history_obj_scatter based on cumulative setting
        if cumulative:
            # Find all indices up to current frame_idx
            plot_reports = indices[: frame_idx + 1]
        else:
            plot_reports = [gen_idx]

        # Plot using the history plotting function
        history_obj_scatter(
            history,
            reports=plot_reports,
            fig=fig,
            ax=ax,
            domination_filt=domination_filt,
            feasibility_filt=feasibility_filt,
            show_points=show_points,
            n_pf=n_pf,
            pf_objectives=pf_objectives,
            show_attainment=show_attainment,
            show_dominated_area=show_dominated_area,
            ref_point=ref_point,
            ref_point_padding=ref_point_padding,
            legend_loc=legend_loc,
            scale=scale,
            flip_objs=flip_objs,
            show_names=show_names,
            show_pf=show_pf,
            single_color=single_color,
            generation_mode="cumulative",
        )

        # Add generation counter
        generation = gen_idx + 1
        fevals = history.reports[gen_idx].fevals
        ax.set_title(f"Generation {generation} (Fevals: {fevals})")

        if n_objectives == 2:
            if not dynamic_scaling:
                # Use global limits
                lim_idx = [1, 0] if flip_objs else [0, 1]
                lims = [ylim, xlim] if flip_objs else [xlim, ylim]
                ax.set_xlim(scale[lim_idx[0]] * lims[0][0], scale[lim_idx[0]] * lims[0][1])
                ax.set_ylim(scale[lim_idx[1]] * lims[1][0], scale[lim_idx[1]] * lims[1][1])

        return (ax,)

    # Create and return the animation
    anim = animation.FuncAnimation(
        fig=fig,
        func=update,
        frames=len(indices),  # Now use length of selected indices
        interval=interval,
        blit=False,
    )

    return anim


def history_dvar_animation(
    history: History,
    reports: Optional[Union[int, slice, List[int], Tuple[int, int]]] = None,
    dvars: Optional[Union[int, slice, List[int], Tuple[int, int]]] = None,
    interval: int = 200,
    domination_filt: Literal["all", "dominated", "non-dominated"] = "all",
    feasibility_filt: Literal["all", "feasible", "infeasible"] = "all",
    hist_bins: Optional[int] = None,
    show_names: bool = True,
    lower_bounds: Optional[np.ndarray] = None,
    upper_bounds: Optional[np.ndarray] = None,
    scale: Optional[np.ndarray] = None,
    single_color: Optional[str] = None,
    plot_bounds: bool = False,
    dynamic_scaling: bool = False,
    cumulative: bool = False,
    scale_padding=0.05,
) -> animation.Animation:
    """
    Creates an animated visualization of how the decision variables evolve across generations.

    Parameters
    ----------
    history : paretobench History
        The history object containing populations with data to plot
    reports : int, slice, List[int], or Tuple[int, int], optional
        Specifies which generations to animate. See `selection_to_indices` for more details.
    dvars : int, slice, List[int], or Tuple[int, int], optional
        Which decision vars to plot. See `population_dvar_pairs` docstring for more details.
    interval : int, optional
        Delay between frames in milliseconds, by default 200
    domination_filt : Literal["all", "dominated", "non-dominated"], optional
        Plot only the dominated/non-dominated solutions, or all. Defaults to all
    feasibility_filt : Literal['all', 'feasible', 'infeasible'], optional
        Plot only the feasible/infeasible solutions, or all. Defaults to all
    hist_bins : int, optional
        Number of bins for histograms on the diagonal, default is to let matplotlib choose
    show_names : bool, optional
        Whether to include variable names on the axes if they exist, default is True
    lower_bounds : array-like, optional
        Lower bounds for each decision variable
    upper_bounds : array-like, optional
        Upper bounds for each decision variable
    scale : array-like, optional
        Scale factors for each variable. Must have the same length as the number of decision vars.
        If None, no scaling is applied.
    single_color: Optional[str] = None
        Color to use when generation_mode is 'cumulative'. If None, uses default color from matplotlib.
    plot_bounds: bool = False
        Whether to plot bounds for the problem
    dynamic_scaling : bool, optional
        If True, axes limits will update based on each frame's data.
        If False, axes limits will be fixed based on all data, by default False
    cumulative : bool, optional
        If True, shows all points seen up to current frame.
        If False, shows only current frame's points, by default False
    scale_padding : float, optional
        Padding used when calculating axis limits w/ dynamic limits

    Returns
    -------
    animation.Animation
        The animation object that can be displayed in notebooks or saved to file
    """
    if not history.reports:
        raise ValueError("No populations in history to animate")

    if scale is not None:
        scale = np.asarray(scale)
        if len(scale.shape) != 1 or scale.shape[0] != history.reports[0].n:
            raise ValueError(
                f"Length of scale must match number of decision vars. Got scale factors with shape {scale.shape}"
                f" and {history.reports[0].x.shape[1]} decision vars."
            )
    else:
        scale = np.ones(history.reports[0].n)

    # Handle different types of selection
    indices = selection_to_indices(reports, len(history.reports))

    if not indices:
        raise ValueError("No reports selected")

    # Create initial plot to get figure and axes using the first selected generation
    fig, axes = history_dvar_pairs(
        history,
        reports=indices[0],
        dvars=dvars,
        domination_filt=domination_filt,
        feasibility_filt=feasibility_filt,
        hist_bins=hist_bins,
        show_names=show_names,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        scale=scale,
        single_color=single_color,
        plot_bounds=plot_bounds,
        generation_mode="cumulative",
    )

    # Calculate global axis limits if not using dynamic scaling
    if not dynamic_scaling:
        # Only use selected generations for calculating limits
        selected_reports = [history.reports[idx] for idx in indices]
        all_x = np.vstack([pop.x for pop in selected_reports])
        n_vars = all_x.shape[1]

        # Calculate limits for each variable
        var_limits = []
        for i in range(n_vars):
            var_min, var_max = np.min(all_x[:, i]), np.max(all_x[:, i])
            limit = (var_min - (var_max - var_min) * scale_padding, var_max + (var_max - var_min) * scale_padding)
            var_limits.append(limit)

    # Function to update frame for animation
    def update(frame_idx):
        # Clear all axes
        for ax in axes.flat:
            ax.clear()

        # Get the actual generation index from our selected indices
        gen_idx = indices[frame_idx]

        # Calculate the appropriate reports for history_dvar_pairs based on cumulative setting
        if cumulative:
            # Find all indices up to current frame_idx
            plot_reports = indices[: frame_idx + 1]
        else:
            plot_reports = [gen_idx]

        # Plot using the history plotting function
        history_dvar_pairs(
            history,
            reports=plot_reports,
            dvars=dvars,
            fig=fig,
            axes=axes,
            domination_filt=domination_filt,
            feasibility_filt=feasibility_filt,
            hist_bins=hist_bins,
            show_names=show_names,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            scale=scale,
            single_color=single_color,
            plot_bounds=plot_bounds,
            generation_mode="cumulative",
        )

        # Add generation counter
        generation = gen_idx + 1
        fevals = history.reports[gen_idx].fevals
        fig.suptitle(f"Generation {generation} (Fevals: {fevals})")

        if not dynamic_scaling:
            # Apply global limits to all subplots including histograms
            for i, ax in enumerate(axes.flat):
                row = i // axes.shape[1]
                col = i % axes.shape[1]
                if row == col:
                    # For histograms on diagonal
                    ax.set_xlim(scale[row] * var_limits[row][0], scale[row] * var_limits[row][1])
                else:
                    # For scatter plots (both upper and lower triangle)
                    ax.set_xlim(scale[col] * var_limits[col][0], scale[col] * var_limits[col][1])
                    ax.set_ylim(scale[row] * var_limits[row][0], scale[row] * var_limits[row][1])

        return tuple(axes.flat)

    # Create and return the animation
    anim = animation.FuncAnimation(
        fig=fig,
        func=update,
        frames=len(indices),  # Now use length of selected indices
        interval=interval,
        blit=False,
    )

    return anim



from .problem import Problem, ProblemWithFixedPF, ProblemWithPF
from .factory import register_problem, get_problem_names, create_problem
from .containers import Population, History, Experiment
from .metrics import InverseGenerationalDistance, eval_metrics_experiments
from .analyze_metrics import (
    aggregate_metrics_feval_budget,
    construct_metric_comparison_table,
    comparison_table_to_latex,
)
from .exceptions import (
    SerializationError,
    DeserializationError,
    UnknownProblemError,
    InputError,
    EmptyPopulationError,
    NoObjectivesError,
    NoConstraintsError,
    NoDecisionVarsError,
)

# Test problems
from .cf import CF1, CF2, CF3, CF4, CF5, CF6, CF7, CF8, CF9, CF10
from .ctp import CTP1, CTP2, CTP3, CTP4, CTP5, CTP6, CTP7
from .dtlz import DTLZ1, DTLZ2, DTLZ3, DTLZ4, DTLZ5, DTLZ6, DTLZ7, DTLZ8, DTLZ9
from .misc import SCH, FON, POL, KUR, CONSTR, SRN, TNK, WATER
from .wfg import WFG1, WFG2, WFG3, WFG4, WFG5, WFG6, WFG7, WFG8, WFG9
from .zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6

_PROBLEM_CLASSES = [
    # DTLZ suite
    *[DTLZ1, DTLZ2, DTLZ3, DTLZ4, DTLZ5, DTLZ6, DTLZ7, DTLZ8, DTLZ9],
    # ZDT suite
    *[ZDT1, ZDT2, ZDT3, ZDT4, ZDT6],
    # WFG suite
    *[WFG1, WFG2, WFG3, WFG4, WFG5, WFG6, WFG7, WFG8, WFG9],
    # Classical problems
    *[SCH, FON, POL, KUR, CONSTR, SRN, TNK, WATER],
    # CTP suite
    *[CTP1, CTP2, CTP3, CTP4, CTP5, CTP6, CTP7],
    # CF suite
    *[CF1, CF2, CF3, CF4, CF5, CF6, CF7, CF8, CF9, CF10],
]

for problem_class in _PROBLEM_CLASSES:
    register_problem(problem_class)

__all__ = [
    "Problem",
    "ProblemWithFixedPF",
    "ProblemWithPF",
    "register_problem",
    "get_problem_names",
    "create_problem",
    "SerializationError",
    "DeserializationError",
    "UnknownProblemError",
    "InputError",
    "Population",
    "History",
    "Experiment",
    "InverseGenerationalDistance",
    "eval_metrics_experiments",
    "aggregate_metrics_feval_budget",
    "construct_metric_comparison_table",
    "comparison_table_to_latex",
    "EmptyPopulationError",
    "NoObjectivesError",
    "NoConstraintsError",
    "NoDecisionVarsError",
] + [cls.__name__ for cls in _PROBLEM_CLASSES]



import numpy as np

from ..containers import Population
from ..utils import get_nondominated_inds


def get_reference_point(population: Population, padding=0.1):
    """Calculate reference point for the attainment surface from all points in the population.

    Parameters
    ----------
    population : Population
        Population to compute reference point for.
    padding : float, optional
        Padding factor to extend reference point beyond maximum values.
        Default is 0.1 (10% extension).

    Returns
    -------
    ndarray
        Reference point coordinates calculated as max values plus padding.
    """
    min_vals = np.min(population.f, axis=0)
    max_vals = np.max(population.f, axis=0)
    ref_point = max_vals + (max_vals - min_vals) * padding
    return ref_point


def compute_attainment_surface_2d(population: Population, ref_point=None, padding=0.1):
    """
    Compute the attainment surface for a set of points in 2D. The surface
    consists of horizontal and vertical lines connecting the points, forming a
    staircase-like pattern.

    Parameters
    ----------
    population : Population
        The population of individuals to compute the attainment surface for.
    ref_point : np.ndarray
        2D reference point for extending surface until
    padding : float
        How much padding to apply if we compute ref_point ourselves

    Returns
    -------
    np.ndarray
        Array of points defining the attainment surface.
    """
    if population.m != 2:
        raise ValueError("Attainment surface can only be computed for 2D points")

    # This only works on feasible individuals
    population = population[population.get_feasible_indices()]

    if len(population) == 0:
        return np.empty((0, 2))

    # Handle missing ref-point
    if ref_point is None:
        ref_point = get_reference_point(population, padding=padding)
    ref_point = np.asarray(ref_point)

    # Get only nondominated points
    population = population.get_nondominated_set()

    if not np.all(ref_point >= np.max(population.f, axis=0)):
        raise ValueError(
            f"Reference point coordinates must exceed all points in non-dominated set "
            f"(ref_point={ref_point}, max_pf=({np.max(population.f[:, 0])}, {np.max(population.f[:, 1])}))"
        )

    # Sort points by x coordinate (first objective)
    sorted_points = population.f[np.argsort(population.f[:, 0])]

    # Initialize the surface points list with the first point
    surface = []
    surface.append(sorted_points[0])

    # Generate horizontal-then-vertical segments between each pair of points
    for i in range(len(sorted_points) - 1):
        current = sorted_points[i]
        next_point = sorted_points[i + 1]

        # Add horizontal line point
        surface.append([next_point[0], current[1]])
        # Add the next point
        surface.append(next_point)
    surface = np.array(surface)
    return np.concatenate(
        (
            [[surface[0, 0], ref_point[1]]],
            surface,
            [[ref_point[0], surface[-1, 1]]],
        ),
        axis=0,
    )


def save_mesh_to_stl(vertices: np.ndarray, triangles: np.ndarray, filename: str):
    """
    Save a triangular mesh to STL file format.

    Args:
        vertices: (n,3) array of vertex coordinates
        triangles: (m,3) array of triangle indices into vertices
        filename: output filename (should end in .stl)
    """
    with open(filename, "w") as f:
        f.write("solid attainment_surface\n")

        # For each triangle
        for triangle in triangles:
            # Get vertex coordinates for this triangle
            v0 = vertices[triangle[0]]
            v1 = vertices[triangle[1]]
            v2 = vertices[triangle[2]]

            # Compute normal using cross product
            edge1 = v1 - v0
            edge2 = v2 - v0
            normal = np.cross(edge1, edge2)
            # Normalize
            length = np.sqrt(np.sum(normal**2))
            if length > 0:
                normal = normal / length

            # Write facet
            f.write(f" facet normal {normal[0]:.6e} {normal[1]:.6e} {normal[2]:.6e}\n")
            f.write("  outer loop\n")
            f.write(f"   vertex {v0[0]:.6e} {v0[1]:.6e} {v0[2]:.6e}\n")
            f.write(f"   vertex {v1[0]:.6e} {v1[1]:.6e} {v1[2]:.6e}\n")
            f.write(f"   vertex {v2[0]:.6e} {v2[1]:.6e} {v2[2]:.6e}\n")
            f.write("  endloop\n")
            f.write(" endfacet\n")


def get_vertex_index(point, vertex_dict, vertices):
    """Helper function to get or create vertex index for a point."""
    point_tuple = tuple(point)
    if point_tuple not in vertex_dict:
        vertex_dict[point_tuple] = len(vertices)
        vertices.append(point)
    return vertex_dict[point_tuple]


def find_rectangles(valid_cells, coords1, coords2):
    """
    Find maximal rectangles in a binary grid.
    Returns list of (min_coord1, min_coord2, max_coord1, max_coord2) for each rectangle.
    """
    if not valid_cells.any():
        return []

    rectangles = []
    remaining = valid_cells.copy()

    while remaining.any():
        # Find first remaining true cell
        row_idx, col_idx = np.nonzero(remaining)
        start_row, start_col = row_idx[0], col_idx[0]

        # Try to expand rectangle right and down
        max_col = start_col
        while max_col + 1 < remaining.shape[1] and remaining[start_row, max_col + 1]:
            max_col += 1

        max_row = start_row
        while max_row + 1 < remaining.shape[0]:
            can_expand = True
            for c in range(start_col, max_col + 1):
                if not remaining[max_row + 1, c]:
                    can_expand = False
                    break
            if not can_expand:
                break
            max_row += 1

        # Add rectangle
        rectangles.append((coords1[start_col], coords2[start_row], coords1[max_col + 1], coords2[max_row + 1]))

        # Mark used cells as processed
        remaining[start_row : max_row + 1, start_col : max_col + 1] = False

    return rectangles


def mesh_plane(sorted_points, fixed_dim, dim1, dim2, reference, vertex_dict, vertices):
    """
    Generate mesh for a plane where fixed_dim is the sorting dimension
    and dim1, dim2 are the dimensions to create the grid in.
    """
    triangles_plane = []

    # Process each point by increasing fixed_dim coordinate
    for i, current_point in enumerate(sorted_points):
        # Get points with smaller fixed_dim coordinate
        previous_points = sorted_points[:i]
        current_fixed = current_point[fixed_dim]

        if len(previous_points) == 0:
            # If no previous points, just add one rectangle up to reference point
            vertices_3d = []
            for p in [
                [current_point[dim1], current_point[dim2]],
                [reference[dim1], current_point[dim2]],
                [current_point[dim1], reference[dim2]],
                [reference[dim1], reference[dim2]],
            ]:
                coord = np.zeros(3)
                coord[fixed_dim] = current_fixed
                coord[dim1] = p[0]
                coord[dim2] = p[1]
                vertices_3d.append(get_vertex_index(coord, vertex_dict, vertices))

            triangles_plane.extend(
                [[vertices_3d[0], vertices_3d[1], vertices_3d[3]], [vertices_3d[0], vertices_3d[3], vertices_3d[2]]]
            )
            continue

        # Get nondominated points from previous layers
        previous_2d = np.column_stack((previous_points[:, dim1], previous_points[:, dim2]))
        nd_points = previous_2d[get_nondominated_inds(previous_2d), :]

        # Get unique coordinates including current point and reference
        coords1 = np.unique(np.concatenate([nd_points[:, 0], [current_point[dim1], reference[dim1]]]))
        coords2 = np.unique(np.concatenate([nd_points[:, 1], [current_point[dim2], reference[dim2]]]))

        # Create grid of valid cells
        valid_cells = np.zeros((len(coords2) - 1, len(coords1) - 1), dtype=bool)

        # For each grid cell
        for row in range(valid_cells.shape[0]):
            for col in range(valid_cells.shape[1]):
                cell_min = np.array([coords1[col], coords2[row]])

                # Check if cell is dominated by current point
                if cell_min[0] >= current_point[dim1] and cell_min[1] >= current_point[dim2]:
                    # Check if cell is not dominated by any previous nondominated point
                    is_dominated_by_previous = False
                    for nd_point in nd_points:
                        if cell_min[0] >= nd_point[0] and cell_min[1] >= nd_point[1]:
                            is_dominated_by_previous = True
                            break

                    valid_cells[row, col] = not is_dominated_by_previous

        # Find maximal rectangles in valid cells
        rectangles = find_rectangles(valid_cells, coords1, coords2)

        # Create triangles for each rectangle
        for rect_min1, rect_min2, rect_max1, rect_max2 in rectangles:
            vertices_3d = []
            for p in [[rect_min1, rect_min2], [rect_max1, rect_min2], [rect_min1, rect_max2], [rect_max1, rect_max2]]:
                coord = np.zeros(3)
                coord[fixed_dim] = current_fixed
                coord[dim1] = p[0]
                coord[dim2] = p[1]
                vertices_3d.append(get_vertex_index(coord, vertex_dict, vertices))

            triangles_plane.extend(
                [[vertices_3d[0], vertices_3d[1], vertices_3d[3]], [vertices_3d[0], vertices_3d[3], vertices_3d[2]]]
            )

    return vertices, triangles_plane


def compute_attainment_surface_3d(population: Population, ref_point=None, padding=0.1):
    """
    Generate triangular mesh for union of cuboids attainment surface.

    Parameters
    ----------
    pop : Population
        The population of individuals to compute the attainment surface for.
    reference : ndarray, shape (3,)
        Array defining the other corner of each cuboid starting from nondominated individuals in the population.

    Returns
    -------
    vertices : ndarray, shape (m, 3)
        Array of unique vertices in the mesh.
    triangles : ndarray, shape (k, 3)
        Array of indices into vertices defining triangles.
    """
    if population.m != 3:
        raise ValueError("This function only works for 3D points")

    # This only works on feasible individuals
    population = population[population.get_feasible_indices()]

    if len(population) == 0:
        return np.empty((0, 3)), np.empty((0,))

    # If no reference point provided, compute one
    if ref_point is None:
        ref_point = get_reference_point(population, padding=padding)
    ref_point = np.asarray(ref_point)

    if not np.all(ref_point >= np.max(population.f, axis=0)):
        raise ValueError("Reference point must dominate all points")

    # Get the nondominated points
    population = population.get_nondominated_set()

    vertices = []
    triangles = []
    vertex_dict = {}

    # Sort points in each dimension
    sorted_by_z = population.f[np.argsort(population.f[:, 2])]  # For XY plane
    sorted_by_x = population.f[np.argsort(population.f[:, 0])]  # For YZ plane
    sorted_by_y = population.f[np.argsort(population.f[:, 1])]  # For XZ plane

    # Process XY plane (sorted by Z)
    triangles.extend(mesh_plane(sorted_by_z, 2, 0, 1, ref_point, vertex_dict, vertices)[1])

    # Process YZ plane (sorted by X)
    triangles.extend(mesh_plane(sorted_by_x, 0, 1, 2, ref_point, vertex_dict, vertices)[1])

    # Process XZ plane (sorted by Y)
    triangles.extend(mesh_plane(sorted_by_y, 1, 0, 2, ref_point, vertex_dict, vertices)[1])

    return np.array(vertices), np.array(triangles)



import numpy as np
from pydantic import BaseModel

from .exceptions import DeserializationError, InputError
from .factory import create_problem
from .simple_serialize import dumps, loads


class Problem(BaseModel):
    """
    The overarching class all problems inherit from. Children must implement the following methods and properties.
     * `m`: property, the number of objectives
     * `n`: property, the number of decision variables
     * `n_constraints`: property, the number of constraints
     * `var_upper_bounds`: property, the array of upper bounds for decision variables
     * `var_lower_bounds`: property, the array of lower bounds for decision variables
     * `_call`: method, accepts `x` the decision variables (first dimension is batch), return `Population` object
    """

    def __call__(self, x: np.ndarray, check_bounds=True):
        """
        Returns the values of the objective functions and constraints at the decision variables `x`.
        The input can be either batched or a single value.

        Note: When subclassing `Problem`, the function must be implemented by defining `_call`, not `__call__`.

        Parameters
        ----------
        x : np.ndarray
            The decision variables. When batched, the first dimension is the batch dimension.
        check_bounds : bool, optional
            Whether or not to check that `x` is within the boundaries of the problem. Defaults to True.

        Returns
        -------
        Population
            A population object containing the objectives and constraints.
        """
        # If a single input was provided
        if len(x.shape) == 1:
            if x.shape[0] != self.n_vars:
                msg = (
                    f"Input does not match number of decision variables (n_vars={self.n_vars}, x.shape[0]={x.shape[0]})"
                )
                raise InputError(msg)
            if check_bounds and ((x > self.var_upper_bounds).all() or (x < self.var_lower_bounds).all()):
                raise InputError("Input lies outside of problem bounds.")
            pop = self._call(x[None, :])

        # If batched input is used
        elif len(x.shape) == 2:
            if x.shape[1] != self.n_vars:
                msg = (
                    f"Input does not match number of decision variables (n_vars={self.n_vars}, x.shape[1]={x.shape[1]})"
                )
                raise InputError(msg)
            if check_bounds and ((x > self.var_upper_bounds).all() or (x < self.var_lower_bounds).all()):
                raise InputError("Input lies outside of problem bounds.")
            pop = self._call(x)

        # If user provided something not usable
        else:
            raise ValueError(f"Incompatible shape of input array x: {x.shape}")

        # Set the decision variables
        pop.x = x
        return pop

    def _call(self, x: np.ndarray):
        """
        This method is implemented by the child classes of `Problem` and should operate on a batched array of inputs.
        """
        raise NotImplementedError()

    @property
    def n_vars(self):
        """
        Returns the number of decision variables expected by this problem. Passed through to property `n`.
        """
        return self.n

    @property
    def n_objs(self):
        """
        Returns the number of objective functions used in this problem. Passed through to property `m`.
        """
        return self.m

    @property
    def n_constraints(self):
        """
        Returns the number of constraints in the problem
        """
        return 0

    @property
    def var_lower_bounds(self):
        """
        Returns the rectangular lower boundaries of the decision variables.
        """
        raise NotImplementedError()

    @property
    def var_upper_bounds(self):
        """
        Returns the rectangular upper boundaries of the decision variables
        """
        raise NotImplementedError()

    @property
    def var_bounds(self):
        """
        Returns the rectangular boundaries of the decision variables (2d numpy array
        where first row is lower bound of each variable and second row are the upper bounds)
        """
        return np.vstack((self.var_lower_bounds, self.var_upper_bounds))

    @property
    def reference(self):
        """
        Returns an APA formatted reference to where the problem was defined.
        """
        raise NotImplementedError()

    def to_line_fmt(self):
        """Serializes the problem object and returns it in a single line human readable format with the problem name and all of
        the data required to recreate it.

        Returns
        -------
        str
            The serialized problem object.
        """
        # Grab problem name and parameters
        name = type(self).__name__
        params = self.model_dump()

        # Save with parameters or just give name if no parameters
        if params:
            return f"{   name } ({ dumps(params) })"
        return name

    @classmethod
    def from_line_fmt(cls, s: str):
        """Create a problem object from the "single line" format. When run from the abstract class `Problem` this expects a
        string of the format `NAME (PARAMETERS)` or `NAME` and will create a problem object of the right class name with the
        specified parameters. If called from a child class, it expects the argument to only contain the paraemeters and creates
        the class based on that.

        Parameters
        ----------
        s : str
            The single line describing the problem object

        Returns
        -------
        Problem
            The instantiated problem

        Raises
        ------
        DeserializationError
            The string couldn't be parsed into the format NAME (PARAMETERS)
        """
        # Run from the abstract class
        if cls == Problem:
            # Find the section of the string corresponding to serialized parameters
            serialization_beg = s.find("(")
            serialization_end = s.find(")")

            # No parameters were passed
            if (serialization_beg == -1) and (serialization_end == -1):
                name = s.strip()
                kwargs = {}
            elif (serialization_beg != -1) and (serialization_end != -1):
                name = s[:serialization_beg].strip()
                kwargs = loads(s[serialization_beg + 1 : serialization_end])
            else:
                raise DeserializationError('could not interpret line "s"')

            # Create the problem and return
            return create_problem(name, **kwargs)

        # We are called from a child class; load parameters and create
        else:
            return cls(**loads(s))

    def __repr__(self):
        return self.to_line_fmt()

    def __str__(self):
        return self.__repr__()


class ProblemWithPF:
    """
    Mixin class for problems with a defined Pareto front where you can request a certain number of points from it.
    """

    def get_pareto_front(self, n=1000):
        """
        Returns at lesat n points along the Pareto front.
        """
        raise NotImplementedError()


class ProblemWithFixedPF:
    """
    Mixin class for problems that have a limited number of points along the Pareto front (ie you can't request a number of them)
    """

    def get_pareto_front(self):
        """
        Returns all of the points on the Pareto front.
        """
        raise NotImplementedError()



import numpy as np

from .problem import Problem
from .containers import Population


class SCH(Problem):
    """
    Description: Convex

    Other name: Schaffer function N. 1
    """

    @property
    def n(self):
        return 1

    @property
    def m(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        return Population(f=np.vstack((x[0] ** 2, (x[0] - 2) ** 2)).T)

    @property
    def var_lower_bounds(self):
        return -1e3 * np.ones(self.n)

    @property
    def var_upper_bounds(self):
        return 1e3 * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class FON(Problem):
    """
    Description: non-convex

    Other name: Fonseca-Fleming Function
    """

    @property
    def n(self):
        return 3

    @property
    def m(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        term1 = -((x[0] - 1 / np.sqrt(3)) ** 2) - (x[1] - 1 / np.sqrt(3)) ** 2 - (x[2] - 1 / np.sqrt(3)) ** 2
        term2 = -((x[0] + 1 / np.sqrt(3)) ** 2) - (x[1] + 1 / np.sqrt(3)) ** 2 - (x[2] + 1 / np.sqrt(3)) ** 2
        return Population(
            f=np.array(
                [
                    1 - np.exp(term1),
                    1 - np.exp(term2),
                ]
            ).T
        )

    @property
    def var_lower_bounds(self):
        return -4 * np.ones(self.n)

    @property
    def var_upper_bounds(self):
        return 4 * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class POL(Problem):
    """
    Description: non-convex, disconnected

    Other name: Poloni’s two objective function
    """

    @property
    def n(self):
        return 2

    @property
    def m(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        a1 = 0.5 * np.sin(1) - 2 * np.cos(1) + np.sin(2) - 1.5 * np.cos(2)
        a2 = 1.5 * np.sin(1) - np.cos(1) + 2 * np.sin(2) - 0.5 * np.cos(2)
        b1 = 0.5 * np.sin(x[0]) - 2 * np.cos(x[0]) + np.sin(x[1]) - 1.5 * np.cos(x[1])
        b2 = 1.5 * np.sin(x[0]) - np.cos(x[0]) + 2 * np.sin(x[1]) - 0.5 * np.cos(x[1])
        return Population(
            f=np.array(
                [
                    1 + (a1 - b1) ** 2 + (a2 - b2) ** 2,
                    (x[0] + 3) ** 2 + (x[1] + 1) ** 2,
                ]
            ).T
        )

    @property
    def var_lower_bounds(self):
        return -np.pi * np.ones(self.n)

    @property
    def var_upper_bounds(self):
        return np.pi * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class KUR(Problem):
    """
    Description: non-convex

    Other name: Kursawe’s Function
    """

    n: int = 3

    @property
    def m(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        return Population(
            f=np.array(
                [
                    np.sum(-10 * np.exp(-0.2 * np.sqrt(x[:-1] ** 2 + x[1:] ** 2)), axis=0),
                    np.sum(np.abs(x) ** 0.8 + 5 * np.sin(x**3), axis=0),
                ]
            ).T
        )

    @property
    def var_lower_bounds(self):
        return -5 * np.ones(self.n)

    @property
    def var_upper_bounds(self):
        return 5 * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class CONSTR(Problem):
    @property
    def n(self):
        return 2

    @property
    def m(self):
        return 2

    @property
    def n_constraints(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        f = np.array([x[0], (1 + x[1]) / x[0]])
        g = np.array([x[1] + 9 * x[0] - 6, -x[1] + 9 * x[0] - 1])
        return Population(f=f.T, g=g.T)

    @property
    def var_lower_bounds(self):
        return np.array([0.1, 0.0])

    @property
    def var_upper_bounds(self):
        return np.array([[1.0, 5.0]])

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class SRN(Problem):
    @property
    def n(self):
        return 2

    @property
    def m(self):
        return 2

    @property
    def n_constraints(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        f = np.array([(x[0] - 2) ** 2 + (x[1] - 1) ** 2 + 2, 9 * x[0] - (x[1] - 1) ** 2])
        g = np.array([225 - (x[0] ** 2 + x[1] ** 2), -10 - (x[0] - 3 * x[1])])
        return Population(f=f.T, g=g.T)

    @property
    def var_lower_bounds(self):
        return -20 * np.ones(self.n)

    @property
    def var_upper_bounds(self):
        return 20 * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class TNK(Problem):
    @property
    def n(self):
        return 2

    @property
    def m(self):
        return 2

    @property
    def n_constraints(self):
        return 2

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        f = np.array([x[0], x[1]])
        g = np.array(
            [
                -(-(x[0] ** 2) - x[1] ** 2 + 1 + 0.1 * np.cos(16 * np.arctan(x[0] / x[1]))),
                0.5 - ((x[0] - 0.5) ** 2 + (x[1] - 0.5) ** 2),
            ]
        )
        return Population(f=f.T, g=g.T)

    @property
    def var_bounds(self):
        return np.array([[0.0, 0.0], [np.pi, np.pi]])

    @property
    def var_lower_bounds(self):
        return np.zeros(self.n)

    @property
    def var_upper_bounds(self):
        return np.pi * np.ones(self.n)

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )


class WATER(Problem):
    @property
    def n(self):
        return 3

    @property
    def m(self):
        return 5

    @property
    def n_constraints(self):
        return 7

    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        f = np.array(
            [
                106780.37 * (x[1] + x[2]) + 61704.67,
                3000.0 * x[0],
                305700 * 2289 * x[1] / (0.06 * 2289) ** 0.65,
                250 * 2289 * np.exp(-39.75 * x[1] + 9.9 * x[2] + 2.74),
                25 * (1.39 / (x[0] * x[1]) + 4940 * x[2] - 80),
            ]
        )
        g = np.array(
            [
                1.0 - (0.00139 / (x[0] * x[1]) + 4.94 * x[2] - 0.08),
                1.0 - (0.000306 / (x[0] * x[1]) + 1.082 * x[2] - 0.0986),
                50000.0 - (12.307 / (x[0] * x[1]) + 49408.25 * x[2] + 4051.02),
                16000.0 - (2.098 / (x[0] * x[1]) + 8046.33 * x[2] - 696.71),
                10000.0 - (2.138 / (x[0] * x[1]) + 7883.39 * x[2] - 705.04),
                2000.0 - (0.417 / (x[0] * x[1]) + 1721.26 * x[2] - 136.54),
                550.0 - (0.164 / (x[0] * x[1]) + 631.13 * x[2] - 54.48),
            ]
        )
        return Population(f=f.T, g=g.T)

    @property
    def var_lower_bounds(self):
        return np.array([0.01, 0.01, 0.01])

    @property
    def var_upper_bounds(self):
        return np.array([0.45, 0.10, 0.10])

    @property
    def reference(self):
        return (
            "Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). A fast and elitist multiobjective genetic "
            "algorithm: NSGA-II. IEEE Transactions on Evolutionary Computation, 6(2), 182–197."
        )



from dataclasses import dataclass
from matplotlib.colors import to_rgb
from typing import Literal, Union, List
import numpy as np

from ..containers import Population
from ..utils import fast_dominated_argsort


@dataclass
class PointSettings:
    nd_inds: np.ndarray
    feas_inds: np.ndarray
    markers: np.ndarray
    plot_filt: np.ndarray
    alpha: np.ndarray


def get_per_point_settings_population(
    population: Population,
    domination_filt: Literal["all", "dominated", "non-dominated"],
    feasibility_filt: Literal["all", "feasible", "infeasible"],
):
    """
    Calculate the per-point settings for scatter plots of the population (ie color, marker, which points are visible)
    based on shared settings across plot types.

    Parameters
    ----------
    population : Population
        Population we are plottings.
    domination_filt : {'all', 'dominated', 'non-dominated'}
        Which points to plot based on domination status.
    feasibility_filt : {'all', 'feasible', 'infeasible'}
        Which points to plot based on feasibility status.

    Returns
    -------
    PointSettings
        Settings object containing:
        nd_inds : ndarray of bool
            Non-dominated indices.
        feas_inds : ndarray of bool
            Feasible indices.
        plot_filt : ndarray of bool
            Which points should be plotted.
        alpha : ndarray of float
            Alpha value per point based on domination rank.
        markers : ndarray of str
            Marker type per point ('o' for feasible, 'x' for infeasible).
    """
    # Break the objectives into those which are non-dominated and those which are not
    nd_inds = population.get_nondominated_indices()
    feas_inds = population.get_feasible_indices()
    markers = np.where(feas_inds, "o", "x")

    # Process filters for what is visible
    plot_filt = np.ones(len(population), dtype=bool)

    # Handle the domination filter
    if domination_filt == "all":
        pass
    elif domination_filt == "dominated":
        plot_filt = np.bitwise_and(plot_filt, ~nd_inds)
    elif domination_filt == "non-dominated":
        plot_filt = np.bitwise_and(plot_filt, nd_inds)
    else:
        raise ValueError(f"Unrecognized option for domination_filt: {domination_filt}")

    # Handle the feasibility filter
    if feasibility_filt == "all":
        pass
    elif feasibility_filt == "feasible":
        plot_filt = np.bitwise_and(plot_filt, feas_inds)
    elif feasibility_filt == "infeasible":
        plot_filt = np.bitwise_and(plot_filt, ~feas_inds)
    else:
        raise ValueError(f"Unrecognized option for feasibility_filt: {feasibility_filt}")

    # Get the domination ranks (of only the visible solutions so we don't end up with a plot of all invisible points)
    ranks = np.zeros(len(population))
    filtered_indices = np.where(plot_filt)[0]
    for rank, idx in enumerate(fast_dominated_argsort(population.f[plot_filt, :], population.g[plot_filt, :])):
        ranks[filtered_indices[idx]] = rank

    # Compute alpha from the ranks
    if np.all(ranks < 1):
        alpha = np.ones(len(population))
    else:
        alpha = 0.5 - ranks / ranks.max() * 0.3
        alpha[ranks == 0] = 1.0

    return PointSettings(nd_inds=nd_inds, feas_inds=feas_inds, plot_filt=plot_filt, alpha=alpha, markers=markers)


def alpha_scatter(ax, x, y, z=None, color=None, alpha=None, marker=None, **kwargs):
    if color is None:
        color = ax._get_lines.get_next_color()

    if alpha is None:
        alpha = 1.0

    # Create a color for each point with the appropriate alpha
    r, g, b = to_rgb(color)
    if isinstance(alpha, float):
        color = (r, g, b, alpha)
    else:
        color = [(r, g, b, a) for a in alpha]

    if marker is None:
        return [ax.scatter(x, y, c=color, **kwargs)]

    if isinstance(marker, str):
        return [ax.scatter(x, y, c=color, marker=marker, **kwargs)]

    # Pull out the label before going into loop
    label = kwargs.pop("label", None)

    # Loop over each possible label in the dataset and plot in batches
    points = []
    unique_markers = set(marker)
    for m in unique_markers:
        mask = np.array(marker) == m
        filtered_color = np.array(color)[mask] if isinstance(color, list) else color
        if z is None:
            points.append(ax.scatter(x[mask], y[mask], c=filtered_color, marker=m, **kwargs))
        else:
            points.append(ax.scatter(x[mask], y[mask], z[mask], c=filtered_color, marker=m, **kwargs))

    # Add an empty scatter object for the legend. Place at end to avoid affecting axis limits
    if label is not None:
        if z is None:
            ax.scatter([], [], color=(r, g, b), label=label, **kwargs)
        else:
            ax.scatter([], [], [], color=(r, g, b), label=label, **kwargs)

    return points


def selection_to_indices(
    selection: Union[None, int, slice, List[int], List[bool], np.ndarray, tuple], arr_len: int
) -> List[int]:
    """Convert various selection formats into a list of positive indices.

    This function handles different ways of selecting elements from an array and converts
    them into a list of valid positive indices within the array bounds.

    The supported types for `selection` are:
    - None: selects all indices
    - int: single index (negative indices count from end)
    - slice: standard Python slice object
    - List[int] or np.ndarray of ints: array of indices (negative indices allowed)
    - List[bool] or np.ndarray of bools: boolean mask where True selects the index
    - tuple: (start, end) pair specifying a range

    Parameters
    ----------
    selection : Union[None, int, slice, List[int], List[bool], np.ndarray, tuple]
        The selection specification. See body of docstring for details.
    arr_len : int
        Length of the array being indexed

    Returns
    -------
    List[int]
        List of valid positive indices corresponding to the selection
    """
    # Convert tuples to slices
    if isinstance(selection, tuple) and len(selection) == 2:
        selection = slice(selection[0], selection[1])

    # Handle different types of selection
    if selection is None:
        # Select all populations
        indices = list(range(arr_len))
    elif isinstance(selection, int) or isinstance(selection, np.integer):
        # Single index - convert negative to positive
        if selection < 0:
            selection = arr_len + selection
        if selection < 0:
            raise IndexError(f"Index {selection} out of range for array with length {arr_len}")
        if selection >= arr_len:
            raise IndexError(f"Index {selection} out of range")
        indices = [selection]
    elif isinstance(selection, slice):
        # Slice - get list of indices
        indices = list(range(*selection.indices(arr_len)))
    elif isinstance(selection, (list, np.ndarray)):
        if len(selection) == 0:
            indices = []
        # Check if it's a boolean array
        elif isinstance(selection[0], bool) or (isinstance(selection, np.ndarray) and selection.dtype == bool):
            # Handle boolean mask
            if len(selection) != arr_len:
                raise ValueError(f"Boolean mask length {len(selection)} does not match array length {arr_len}")
            indices = [i for i, selected in enumerate(selection) if selected]
        else:
            # List of indices - convert negative to positive
            indices = []
            for i in selection:
                if not isinstance(i, (int, np.integer)):
                    raise TypeError(f"Invalid index type {type(i)}. Expected integer.")
                idx = int(i) if i >= 0 else arr_len + int(i)
                if idx < 0 or idx >= arr_len:
                    raise IndexError(f"Index {i} out of range for array with length {arr_len}")
                indices.append(idx)
    else:
        raise ValueError(f"Unsupported selection type: {type(selection)}")

    return indices



import numpy as np


def g_1_3(x, k):
    x -= 0.5
    return 100 * (k + np.sum(x**2 - np.cos(20 * np.pi * x), axis=0))


def g_2_4_5(x):
    return np.sum((x - 0.5) ** 2, axis=0)


def theta_5_6(x, g, m):
    th = (1 + 2 * g[None, :] * x[: m - 1, :]) / (2 + 2 * g[None, :])
    return np.vstack([x[0], th[1:, :]])


def f_2_to_6(x, m, alpha=1):
    f1 = np.vstack([np.prod(np.cos(x[: x.shape[0] - i, :] ** alpha * np.pi / 2), axis=0) for i in range(0, m)])
    f2 = np.vstack([np.ones(x.shape[1])] + [np.sin(x[x.shape[0] - i, :] ** alpha * np.pi / 2) for i in range(1, m)])
    return f1 * f2



import numpy as np

from .problem import Problem, ProblemWithPF
from .utils import get_hyperplane_points, get_nondominated_inds
from .wfg_utils import (
    shape_linear,
    shape_convex,
    shape_concave,
    shape_mixed,
    shape_disconnected,
    compute_x,
    transform_bias_polynomial,
    transform_bias_flat_region,
    transform_bias_parameter_dependent,
    transform_shift_linear,
    transform_shift_deceptive,
    transform_shift_multimodal,
    transform_reduction_weighted_sum,
    transform_reduction_non_separable,
)
from .containers import Population


class WFGx(Problem, ProblemWithPF):
    n: int = 24
    k: int = 4
    m: int = 2

    def __init__(self, **data):
        # Handle pydantic data
        super().__init__(**data)

        # Set internal attributes for later
        self._d = 1.0
        self._s = np.array([2 * (i + 1) for i in range(self.m)])

    @property
    def var_lower_bounds(self):
        return np.zeros(self.n)

    @property
    def var_upper_bounds(self):
        return np.array([2 * (i + 1) for i in range(self.n)])

    @property
    def reference(self):
        return (
            "Huband, S., Hingston, P., Barone, L., & While, L. (2006). A review of multiobjective test problems and a scalable test "
            "problem toolkit. IEEE Transactions on Evolutionary Computation, 10(5), 477–506. https://doi.org/10.1109/TEVC.2005.861417"
        )


class WFG1(WFGx):
    def _call(self, x, set_pos_zero=False):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        t1[self.k :] = transform_shift_linear(y[self.k :], 0.35)

        # Transition 2
        t2 = np.copy(t1)
        t2[self.k :] = transform_bias_flat_region(t1[self.k :], 0.8, 0.75, 0.85)
        if set_pos_zero:  # Because of some floating point round-off error issues
            t2[self.k :] = 0.0

        # Transition 3
        t3 = transform_bias_polynomial(np.clip(t2, 0.0, 100.0), 0.02)

        # Transition 4
        w = b
        t4 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t4.append(transform_reduction_weighted_sum(t3[start:end, :], w[start:end]))
        t4.append(transform_reduction_weighted_sum(t3[self.k :, :], w[self.k :]))
        t4 = np.vstack(t4)

        # Find x
        new_x = compute_x(t4, False)

        # Shape functions
        h = shape_convex(new_x)
        h[-1] = shape_mixed(new_x, 5, 1)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n, n_solver_guess=128, n_solver_iter=32):
        # Invert the convex shape function to find points along the reference vectors
        # f1 = (1-cos(x0))(1-cos(x1))(1-cos(x2))...(1-cos(x{n-1}))
        # fi = (1-cos(x0))(1-cos(x1))(1-cos(x2))...(1-sin(x{n-i+1}))
        # fm = (1-sin(x0))
        # Note that: fi/f1 = (1-sin(x{n-i+1}))/(1-cos(x{n-i+1}))  /  ((1-cos(x))(1-cos(x))... for remaining terms after n-i+1
        # Use (1-sin(x))/(1-cos(x)) = y is solved by x = 2*arctan((sqrt(2y) - 1)/(2y-1))
        r = get_hyperplane_points(self.m, n) + 1e-12  # Hack to avoid some numerical issues
        theta = np.zeros_like(r)
        for j in range(1, r.shape[0]):
            t = r[j] * np.prod(1 - np.cos(theta[-j:-1]), axis=0)
            theta[-j - 1] = 2 * np.arctan2((np.sqrt(2 * r[0] * t) - r[0]), (2 * t - r[0])) % (np.pi)
        x = theta / np.pi * 2

        # Solve for the value of x0 which gives the correct fm/f{m-1} where fm is shape_mixed(...) and f{m-1} is from shape_convex(...)
        t = r[-1] / r[-2] * (1 - np.sin(np.pi / 2 * x[1]))

        def solve_fun(x):
            return x + np.cos(10 * np.pi * x + np.pi / 2) / 10 / np.pi - 1 + t[None, :] * (1 - np.cos(np.pi / 2 * x))

        def solve_fun_der(x):
            return 1 - np.sin(10 * np.pi * x + np.pi / 2) + np.pi / 2 * t[None, :] * np.sin(np.pi / 2 * x)

        # Make an initial guess
        x_guess = np.linspace(0, 1, n_solver_guess)
        x[0] = x_guess[np.argmin(np.abs(solve_fun(x_guess[:, None])), axis=0)]

        # Newton Raphson iterations
        for _ in range(n_solver_iter):
            fun_prime = solve_fun_der(x[0][None, :])
            fun_prime[fun_prime == 0.0] = 1e6  # Avoid divide by zero
            x[0] = np.clip(x[0] - solve_fun(x[0][None, :]) / fun_prime, 0, 1)

        # Evaluate the objective functions
        f = shape_convex(x)
        f[-1] = shape_mixed(x, 5, 1)
        return (f * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG2(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        t1[self.k :] = np.clip(transform_shift_linear(y[self.k :], 0.35), 0.0, 1.0)
        # Transition 2
        t2 = np.empty((self.k + (self.n - self.k) // 2, t1.shape[1]))
        t2[: self.k] = t1[: self.k]
        for i in range(self.k + 1, self.k + (self.n - self.k) // 2 + 1):
            # Calculate indices
            idx1 = self.k + 2 * (i - self.k) - 2
            idx2 = self.k + 2 * (i - self.k) - 1

            # Stack the transformed values
            values = np.vstack((t1[idx1], t1[idx2]))

            # Apply transformation and clipping
            t2[i - 1] = np.clip(
                transform_reduction_non_separable(values, 2),
                0.0,
                1.0,
            )

        # Transition 3
        t3 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t3.append(np.clip(transform_reduction_weighted_sum(t2[start:end, :]), 0.0, 1.0))
        t3.append(np.clip(transform_reduction_weighted_sum(t2[self.k :, :]), 0.0, 1.0))
        t3 = np.vstack(t3)

        # Find x
        new_x = compute_x(t3, False)

        # Shape functions
        h = np.empty((self.m, x.shape[1]))
        h[:-1, :] = shape_convex(new_x)[:-1, :]
        h[-1] = shape_disconnected(new_x, 5, 1.0, 1.0)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n, n_solver_guess=2048, n_solver_iter=32):
        # Invert the convex shape function to find points along the reference vectors
        # f1 = (1-cos(x0))(1-cos(x1))(1-cos(x2))...(1-cos(x{n-1}))
        # fi = (1-cos(x0))(1-cos(x1))(1-cos(x2))...(1-sin(x{n-i+1}))
        # fm = (1-sin(x0))
        # Note that: fi/f1 = (1-sin(x{n-i+1}))/(1-cos(x{n-i+1}))  /  ((1-cos(x))(1-cos(x))... for remaining terms after n-i+1
        # Use (1-sin(x))/(1-cos(x)) = y is solved by x = 2*arctan((sqrt(2y) - 1)/(2y-1))
        r = get_hyperplane_points(self.m, 2 * n) + 1e-12  # Hack to avoid some numerical issues
        theta = np.zeros_like(r)
        for j in range(1, r.shape[0]):
            t = r[j] * np.prod(1 - np.cos(theta[-j:-1]), axis=0)
            theta[-j - 1] = 2 * np.arctan2((np.sqrt(2 * r[0] * t) - r[0]), (2 * t - r[0])) % (np.pi)
        x = theta / np.pi * 2

        # Solve for the value of x0 which gives the correct fm/f{m-1} where fm is shape_mixed(...) and f{m-1} is from shape_convex(...)
        t = r[-1] / r[-2] * (1 - np.sin(np.pi / 2 * x[1]))

        def solve_fun(x):
            return x * np.cos(5 * np.pi * x) ** 2 - 1 + t[None, :] * (1 - np.cos(np.pi / 2 * x))

        def solve_fun_der(x):
            return (
                np.cos(5 * np.pi * x) ** 2
                - 10 * np.pi * x * np.cos(5 * np.pi * x) * np.sin(5 * np.pi * x)
                + np.pi / 2 * t[None, :] * np.sin(np.pi / 2 * x)
            )

        # Make an initial guess (biased toward small x)
        x_guess = np.linspace(0, 1, n_solver_guess)
        x[0] = x_guess[np.min(np.argsort(np.abs(solve_fun(x_guess[:, None])), axis=0)[:8], axis=0)]

        # Newton Raphson iterations
        for _ in range(n_solver_iter):
            fun_prime = solve_fun_der(x[0][None, :])
            fun_prime[fun_prime == 0.0] = 1e6  # Avoid divide by zero
            x[0] = np.clip(x[0] - solve_fun(x[0][None, :]) / fun_prime, 0, 1)

        # Evaluate the objective functions
        f = shape_convex(x)
        f[-1] = shape_disconnected(x, 5, 1, 1)
        f = f[:, get_nondominated_inds(f.T)]
        return (f * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG3(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        t1[self.k :] = np.clip(transform_shift_linear(y[self.k :], 0.35), 0.0, 1.0)

        # Transition 2
        t2 = np.empty((self.k + (self.n - self.k) // 2, t1.shape[1]))
        t2[: self.k] = t1[: self.k]
        for i in range(self.k + 1, self.k + (self.n - self.k) // 2 + 1):
            # Calculate indices
            idx1 = self.k + 2 * (i - self.k) - 2
            idx2 = self.k + 2 * (i - self.k) - 1

            # Stack the transformed values
            values = np.vstack((t1[idx1], t1[idx2]))

            # Apply transformation and clipping
            t2[i - 1] = np.clip(
                transform_reduction_non_separable(values, 2),
                0.0,
                1.0,
            )

        # Transition 3
        t3 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t3.append(np.clip(transform_reduction_weighted_sum(t2[start:end, :]), 0.0, 1.0))
        t3.append(np.clip(transform_reduction_weighted_sum(t2[self.k :, :]), 0.0, 1.0))
        t3 = np.vstack(t3)

        # Find x
        new_x = compute_x(t3, True)

        # Shape functions
        h = shape_linear(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = np.vstack((np.linspace(0, 1, n), np.full((self.m - 2, n), 1 / 2), np.zeros((1, n))))
        f = shape_linear(f)
        return (f * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG4(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.clip(transform_shift_multimodal(y, 30, 10, 0.35), 0.0, 1.0)

        # Transition 2
        t2 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t2.append(np.clip(transform_reduction_weighted_sum(t1[start:end, :]), 0.0, 1.0))
        t2.append(np.clip(transform_reduction_weighted_sum(t1[self.k :, :]), 0.0, 1.0))
        t2 = np.vstack(t2)

        # Find x
        new_x = compute_x(t2, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG5(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.clip(transform_shift_deceptive(y, 0.35, 0.001, 0.05), 0.0, 1.0)

        # Transition 2
        t2 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t2.append(np.clip(transform_reduction_weighted_sum(t1[start:end, :]), 0.0, 1.0))
        t2.append(np.clip(transform_reduction_weighted_sum(t1[self.k :, :]), 0.0, 1.0))
        t2 = np.vstack(t2)

        # Find x
        new_x = compute_x(t2, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG6(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1 - checked
        t1 = np.copy(y)
        t1[self.k :] = transform_shift_linear(y[self.k :], 0.35)

        # Transition 2
        gap = self.k // (self.m - 1)
        t2 = [transform_reduction_non_separable(t1[(m - 1) * gap : (m * gap), :], gap) for m in range(1, self.m)]
        t2.append(transform_reduction_non_separable(t1[self.k :, :], self.n - self.k))
        t2 = np.vstack(t2)

        # Find x
        new_x = compute_x(t2, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG7(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        for i in range(1, self.k + 1):
            weighted_sum = transform_reduction_weighted_sum(y[i:, :])
            t1[i - 1] = np.clip(
                transform_bias_parameter_dependent(y[i - 1], weighted_sum, 0.98 / 49.98, 0.02, 50), 0.0, 1.0
            )

        # Transition 2
        t2 = np.copy(t1)
        t2[self.k :] = np.clip(transform_shift_linear(t1[self.k :], 0.35), 0.0, 1.0)

        # Transition 3
        t3 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t3.append(np.clip(transform_reduction_weighted_sum(t2[start:end, :]), 0.0, 1.0))
        t3.append(np.clip(transform_reduction_weighted_sum(t2[self.k :, :]), 0.0, 1.0))
        t3 = np.vstack(t3)

        # Find x
        new_x = compute_x(t3, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG8(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        for i in range(self.k + 1, self.n + 1):
            weighted_sum = transform_reduction_weighted_sum(y[: i - 1])
            t1[i - 1] = np.clip(
                transform_bias_parameter_dependent(y[i - 1][None, :], weighted_sum, 0.98 / 49.98, 0.02, 50),
                0.0,
                1.0,
            )

        # Transition 2
        t2 = np.copy(t1)
        t2[self.k :] = np.clip(transform_shift_linear(t2[self.k :], 0.35), 0.0, 1.0)

        # Transition 3
        t3 = []
        for i in range(1, self.m):
            start = int((i - 1) * self.k / (self.m - 1))
            end = int(i * self.k / (self.m - 1))
            t3.append(np.clip(transform_reduction_weighted_sum(t2[start:end, :]), 0.0, 1.0))
        t3.append(np.clip(transform_reduction_weighted_sum(t2[self.k :, :]), 0.0, 1.0))
        t3 = np.vstack(t3)

        # Find x
        new_x = compute_x(t3, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T


class WFG9(WFGx):
    def _call(self, x):
        # Transpose x (this function was written before ParetoBench standardized on rows being the batched index)
        x = x.T

        # Normalize the input
        b = np.array([2 * (i + 1) for i in range(self.n)])
        y = x / b[:, None]

        # Transition 1
        t1 = np.copy(y)
        for i in range(1, self.n):
            weighted_sum = transform_reduction_weighted_sum(y[i:])
            t1[i - 1] = np.clip(
                transform_bias_parameter_dependent(y[i - 1][None, :], weighted_sum, 0.98 / 49.98, 0.02, 50),
                0.0,
                1.0,
            )

        # Transition 2
        t2 = np.empty_like(t1)
        t2[: self.k] = np.clip(transform_shift_deceptive(t1[: self.k], 0.35, 0.001, 0.05), 0.0, 1.0)
        t2[self.k :] = np.clip(transform_shift_multimodal(t1[self.k :], 30, 95, 0.35), 0.0, 1.0)

        # Transition 3
        gap = self.k // (self.m - 1)
        t3 = [transform_reduction_non_separable(t2[(m - 1) * gap : (m * gap), :], gap) for m in range(1, self.m)]
        t3.append(transform_reduction_non_separable(t2[self.k :, :], self.n - self.k))
        t3 = np.vstack(t3)

        # Find x
        new_x = compute_x(t3, True)

        # Shape functions
        h = shape_concave(new_x)

        # Compute the objective
        obj = (self._d * new_x[-1])[None, :] + self._s[:, None] * h

        # Return it
        return Population(f=obj.T)

    def get_pareto_front(self, n):
        f = get_hyperplane_points(self.m, n)
        return (f / np.sqrt(np.sum(f**2, axis=0)) * 2 * np.arange(1, self.m + 1)[:, None]).T



from itertools import combinations, chain, count
from math import comb
from typing import Union
import numpy as np

from .problem import Problem


def get_betas(m, p):
    """
    From: Das, Indraneel, and J. E. Dennis. “Normal-Boundary Intersection: A New Method for
    Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems.” SIAM
    Journal on Optimization 8, no. 3 (August 1998): 631–57. https://doi.org/10.1137/S1052623496307510.
    """
    beta = np.fromiter(chain.from_iterable(combinations(range(1, p + m), m - 1)), np.float64)
    beta = beta.reshape(beta.shape[0] // (m - 1), m - 1).T
    beta = beta - np.arange(0, m - 1)[:, None] - 1
    beta1 = np.concatenate((beta, np.full((1, beta.shape[1]), p)), axis=0)
    beta2 = np.concatenate((np.zeros((1, beta.shape[1])), beta), axis=0)
    return (beta1 - beta2) / p


def get_hyperplane_points(m, n):
    """
    Returns at least n points of dimension m on the hyperplane x1 + x2 + x3 + ... = 1

    From: Das, Indraneel, and J. E. Dennis. “Normal-Boundary Intersection: A New Method for
    Generating the Pareto Surface in Nonlinear Multicriteria Optimization Problems.” SIAM
    Journal on Optimization 8, no. 3 (August 1998): 631–57. https://doi.org/10.1137/S1052623496307510.
    """
    return get_betas(m, next(p for p in count() if comb(p + m - 1, m - 1) >= n))


def uniform_grid(n, m):
    """
    At lesat n evenly spread points in the hypercube of dimension m
    """
    return np.reshape(
        np.stack(np.meshgrid(*([np.linspace(0, 1, int(np.ceil(n ** (1 / m))))] * m))),
        (m, -1),
    )


def get_domination(objs, constraints=None):
    # Compare all pairs of individuals based on domination
    dom = np.bitwise_and(
        (objs[:, None, :] <= objs[None, :, :]).all(axis=-1),
        (objs[:, None, :] < objs[None, :, :]).any(axis=-1),
    )

    if constraints is not None:
        # If one individual is feasible and the other isn't, set domination
        feas = constraints >= 0.0
        ind = np.bitwise_and(feas.all(axis=1)[:, None], ~feas.all(axis=1)[None, :])
        dom[ind] = True
        ind = np.bitwise_and(~feas.all(axis=1)[:, None], feas.all(axis=1)[None, :])
        dom[ind] = False

        # If both are infeasible, then the individual with the least constraint violation wins
        constraint_violation = -np.sum(np.minimum(constraints, 0), axis=1)
        comp = constraint_violation[:, None] < constraint_violation[None, :]
        ind = ~np.bitwise_or(feas.all(axis=1)[:, None], feas.all(axis=1)[None, :])
        dom[ind] = comp[ind]
    return dom


def fast_dominated_argsort(objs, constraints=None):
    """
    Performs a dominated sort on matrix of objective function values O.  This is a numpy implementation of the algorithm
    described in Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. (2002). IEEE Transactions on Evolutionary
    Computation, 6(2), 182–197. https://doi.org/10.1109/4235.996017

    A list of ranks is returned referencing each individual by its index in the objective matrix.

    :param O: (M, N) numpy array where N is the number of individuals and M is the number of objectives
    :return: List of ranks where each rank is a list of the indices to the individuals in that rank
    """
    # Compare all pairs of individuals based on domination
    dom = get_domination(objs, constraints)

    # Create the sets of dominated individuals, domination number, and first rank
    S = [np.nonzero(row)[0].tolist() for row in dom]
    N = np.sum(dom, axis=0)
    F = [np.where(N == 0)[0].tolist()]

    i = 0
    while len(F[-1]) > 0:
        Q = []
        for p in F[i]:
            for q in S[p]:
                N[q] -= 1
                if N[q] == 0:
                    Q.append(q)
        F.append(Q)
        i += 1

    # Remove last empty set
    F.pop()

    return F


def get_nondominated_inds(objs, constraints=None):
    """
    Returns the indices of the nondominated individuals for the objectives O, an (m,n) array for the m objectives
    """
    dom = get_domination(objs, constraints)
    return np.where(np.sum(dom, axis=0) == 0)[0].tolist()


def triangle_grid_count(n):
    return n * (n + 1) // 2


def triangle_grid(n):
    # skewed points to help compensate for
    x = np.concatenate([np.linspace(0, 1 - np.sqrt(i / (n - 1)), n - i) for i in range(n)])
    y = np.concatenate([np.ones(n - i) * np.sqrt(i / (n - 1)) for i in range(n)])
    return np.vstack((x, y))


def rastrigin(x):
    """
    Compute the Rastragin function for optimization.  The bounds for optimization are [-5.12, 5.12] in all variables

    Reference: Huband, S., Hingston, P., Barone, L., & While, L. (2006). A review of multiobjective test problems and a
    scalable test problem toolkit. IEEE Transactions on Evolutionary Computation, 10(5), 477–506.
    https://doi.org/10.1109/TEVC.2005.861417

    :param x: (m,n) input for m dimensions and n separate instances
    :return: g(x), the Rastrigin function
    """
    a = 10
    return 1 + a * x.shape[0] + np.sum(x**2 - a * np.cos(2 * np.pi * x), axis=0)


def weighted_chunk_sizes(n, weights):
    """
    Break the number n into a list of numbers that sum to n, with the given
    approximate weights

    Args:
        n (int): The number to break apart
        weights (list): list of the weights of each chunk (should sum to one)

    Returns:
        list: The size of each chunk summing to n
    """
    ns = [int(np.floor(n * w / sum(weights))) for w in weights]
    for i in range(32):
        if sum(ns) < n:
            ns[i % len(ns)] += 1
        else:
            break
    return ns


def get_problem_from_obj_or_str(obj_or_str: Union[str, Problem]) -> Problem:
    """Convert input to Problem instance.

    Parameters
    ----------
    obj_or_str : Problem or str
        Input to convert. If already a Problem instance, returns as-is.
        If string, creates Problem from line format.

    Returns
    -------
    Problem
        The resulting Problem instance.

    Raises
    ------
    ValueError
        If input is neither Problem nor str type.
    """
    if isinstance(obj_or_str, Problem):
        return obj_or_str
    elif isinstance(obj_or_str, str):
        return Problem.from_line_fmt(obj_or_str)
    else:
        raise ValueError(f"Unrecognized input type: {type(obj_or_str)}")
